"""The ``add_staff_relationships.py`` script automates adding
the relationships of some staff-related symbols to staffs.
"""
import argparse
import collections
import logging
import os
import pprint
import time

from typing import List
from collections import defaultdict
from mung.constants import InferenceEngineConstants as _CONST
from mung.io import read_nodes_from_file, export_node_list
from mung.node import link_nodes, Node


def add_staff_relationships(nodes: List[Node],
                            notehead_staffspace_threshold: float = 0.2) -> List[Node]:
    id_to_node_mapping = {node.id: node for node in nodes}

    ON_STAFFLINE_RATIO_THRESHOLD = notehead_staffspace_threshold

    ##########################################################################
    logging.info('Find the staff-related symbols')
    staffs = [c for c in nodes if c.class_name == _CONST.STAFF_CLASS_NAME]

    staff_related_symbols = defaultdict(list)  # type: defaultdict[str, List[Node]]
    notehead_symbols = defaultdict(list)  # type: defaultdict[str, List[Node]]
    rest_symbols = defaultdict(list)  # type: defaultdict[str, List[Node]]
    for node in nodes:
        if node.class_name in _CONST.STAFF_RELATED_CLASS_NAMES:
            staff_related_symbols[node.class_name].append(node)
        if node.class_name in _CONST.NOTEHEAD_CLASS_NAMES:
            notehead_symbols[node.class_name].append(node)
        if node.class_name in _CONST.REST_CLASS_NAMES:
            rest_symbols[node.class_name].append(node)

    ##########################################################################
    logging.info('Adding staff relationships')
    #  - Which direction do the relationships lead in?
    #    Need to define this.
    #
    # Staff -> symbol?
    # Symbol -> staff?
    # It does not really matter, but it's more intuitive to attach symbols
    # onto a pre-existing staff. So, symbol -> staff.
    for class_name, nodes in list(staff_related_symbols.items()):
        for node in nodes:  # type: Node
            # Find the related staff. Relatedness is measured by row overlap.
            # That means we have to modify the staff bounding box to lead
            # from the leftmost to the rightmost column. This holds
            # especially for the staff_grouping symbols.
            for staff in staffs:
                top, left, bottom, right = staff.bounding_box
                left = 0
                right = max(right, node.right)
                if node.overlaps((top, left, bottom, right)):
                    link_nodes(node, staff)

    ##########################################################################
    logging.info('Adding rest --> staff relationships.')
    for class_name, nodes in list(rest_symbols.items()):
        for node in nodes:  # type: Node
            closest_staff = min([s for s in staffs],
                                key=lambda x: ((x.bottom + x.top) / 2. - (node.bottom + node.top) / 2.) ** 2)
            link_nodes(node, closest_staff)

    ##########################################################################
    logging.info('Adding notehead relationships.')

    # NOTE:
    # This part should NOT rely on staffspace masks in any way!
    # They are highly unreliable.

    # Sort the staff objects top-down. Assumes stafflines do not cross,
    # and that there are no crazy curves at the end that would make the lower
    # stafflines stick out over the ones above them...
    stafflines = [c for c in nodes if c.class_name == _CONST.STAFFLINE_CLASS_NAME]
    stafflines = sorted(stafflines, key=lambda c: c.top)
    staffspaces = [c for c in nodes if c.class_name == _CONST.STAFFSPACE_CLASS_NAME]
    staffspaces = sorted(staffspaces, key=lambda c: c.top)
    staves = [c for c in nodes if c.class_name == _CONST.STAFF_CLASS_NAME]
    staves = sorted(staves, key=lambda c: c.top)

    # Indexing data structures.
    #
    # We need to know:
    #  - per staffline and staffspace: its containing staff
    _staff_per_ss_sl = {}
    #  - per staffline and staffspace: its index (top to bottom) within the staff
    _ss_sl_idx_wrt_staff = {}
    # Reverse indexes:
    # If I know which staff (by id) and which index of staffline/staffspace,
    # I want to retrieve the given staffline/staffspace Node:
    _staff_and_idx2ss = defaultdict(dict)
    _staff_and_idx2sl = defaultdict(dict)

    # Build the indexes
    for _staff in staves:
        # Keep the top-down ordering from above:
        _s_stafflines = [_staffline for _staffline in stafflines
                         if _staff.id in _staffline.inlinks]
        _s_staffspaces = [_staffspace for _staffspace in staffspaces
                          if _staff.id in _staffspace.inlinks]
        for i, _sl in enumerate(_s_stafflines):
            _staff_per_ss_sl[_sl.id] = _staff
            _ss_sl_idx_wrt_staff[_sl.id] = i
            _staff_and_idx2sl[_staff.id][i] = _sl
            logging.debug('Staff {0}: stafflines {1}'.format(_staff.id,
                                                             _staff_and_idx2sl[_staff.id]))
        for i, _ss in enumerate(_s_staffspaces):
            _staff_per_ss_sl[_ss.id] = _staff
            _ss_sl_idx_wrt_staff[_ss.id] = i
            _staff_and_idx2ss[_staff.id][i] = _ss

    logging.debug(pprint.pformat(dict(_staff_and_idx2ss)))

    for class_name, nodes in list(notehead_symbols.items()):
        for node in nodes:

            ct, cl, cb, cr = node.bounding_box

            ################
            # Add relationship to given staffline or staffspace.

            # If notehead has leger lines, skip it for now.
            has_leger_line = False
            for o in node.outlinks:
                if id_to_node_mapping[o].class_name == _CONST.LEGER_LINE_CLASS_NAME:
                    has_leger_line = True
                    break

            if has_leger_line:
                # Attach to the appropriate staff:
                # meaning, staff closest to the innermost leger line.
                lls = [id_to_node_mapping[o] for o in node.outlinks
                       if id_to_node_mapping[o].class_name == _CONST.LEGER_LINE_CLASS_NAME]
                # Furthest from notehead's top is innermost.
                # (If notehead is below staff and crosses a ll., one
                #  of these numbers will be negative. But that doesn't matter.)
                ll_max_dist = max(lls, key=lambda ll: ll.top - node.top)
                # Find closest staff to max-dist leger ine
                staff_min_dist = min(staves,
                                     key=lambda ss: min((ll_max_dist.bottom - ss.top) ** 2,
                                                        (ll_max_dist.top - ss.bottom) ** 2))
                link_nodes(node, staff_min_dist)
                continue

            # - Find the related staffline.
            # - Because of curved stafflines, this has to be done w.r.t.
            #   the horizontal position of the notehead.
            # - Also, because stafflines are NOT filled in (they do not have
            #   intersections annotated), it is necessary to use a wider
            #   window than just the notehead.
            # - We will assume that STAFFLINES DO NOT CROSS.
            #   (That is a reasonable assumption.)
            #
            # - For now, we only work with more or less straight stafflines.

            overlapped_stafflines = []
            overlapped_staffline_idxs = []
            for i, staff in enumerate(stafflines):
                # This is the assumption of straight stafflines!
                if (ct <= staff.top <= cb) or (ct <= staff.bottom <= cb):
                    overlapped_stafflines.append(staff)
                    overlapped_staffline_idxs.append(i)

            if node.id < 10:
                logging.debug('Notehead {0} ({1}): overlaps {2} stafflines'.format(node.id,
                                                                                   node.bounding_box,
                                                                                   len(overlapped_stafflines), ))

            if len(overlapped_stafflines) == 1:
                staff = overlapped_stafflines[0]
                dtop = staff.top - ct
                dbottom = cb - staff.bottom
                if min(dtop, dbottom) / max(dtop, dbottom) < ON_STAFFLINE_RATIO_THRESHOLD:
                    logging.info('Notehead {0}, staffline {1}: very small ratio {2:.2f}'
                                 ''.format(node.id, staff.id,
                                           min(dtop, dbottom) / max(dtop, dbottom)))
                    # Staffspace?
                    #
                    # To get staffspace:
                    #  - Get orientation (below? above?)
                    _is_staffspace_above = False
                    if dtop > dbottom:
                        _is_staffspace_above = True

                    #  - Find staffspaces adjacent to the overlapped staffline.
                    # NOTE: this will fail with single-staffline staves, because
                    #       they do NOT have the surrounding staffspaces defined...
                    _staffline_idx_wrt_staff = _ss_sl_idx_wrt_staff[staff.id]
                    if _is_staffspace_above:
                        _staffspace_idx_wrt_staff = _staffline_idx_wrt_staff
                    else:
                        _staffspace_idx_wrt_staff = _staffline_idx_wrt_staff + 1

                    # Retrieve the given staffsapce
                    _staff = _staff_per_ss_sl[staff.id]
                    tgt_staffspace = _staff_and_idx2ss[_staff.id][_staffspace_idx_wrt_staff]
                    # Link to staffspace
                    link_nodes(node, tgt_staffspace)
                    # And link to staff
                    _c_staff = _staff_per_ss_sl[tgt_staffspace.id]
                    link_nodes(node, _c_staff)

                else:
                    # Staffline!
                    link_nodes(node, staff)
                    # And staff:
                    _c_staff = _staff_per_ss_sl[staff.id]
                    link_nodes(node, _c_staff)

            elif len(overlapped_stafflines) == 0:
                # Staffspace!
                # Link to the staffspace with which the notehead has
                # greatest vertical overlap.
                #
                # Interesting corner case:
                # Sometimes noteheads "hang out" of the upper/lower
                # staffspace, so they are not entirely covered.
                overlapped_staffspaces = {}
                for _ss_i, staff in enumerate(staffspaces):
                    if staff.top <= node.top <= staff.bottom:
                        overlapped_staffspaces[_ss_i] = min(staff.bottom, node.bottom) - node.top
                    elif node.top <= staff.top <= node.bottom:
                        overlapped_staffspaces[_ss_i] = staff.bottom - max(node.top, staff.top)

                if len(overlapped_staffspaces) == 0:
                    logging.warning('Notehead {0}: no overlapped staffline object, no leger line!'
                                    ''.format(node.id))
                _ss_i_max = max(list(overlapped_staffspaces.keys()),
                                key=lambda x: overlapped_staffspaces[x])
                max_overlap_staffspace = staffspaces[_ss_i_max]
                link_nodes(node, max_overlap_staffspace)
                _c_staff = _staff_per_ss_sl[max_overlap_staffspace.id]
                link_nodes(node, _c_staff)

            elif len(overlapped_stafflines) == 2:
                # Staffspace between those two lines.
                s1 = overlapped_stafflines[0]
                s2 = overlapped_stafflines[1]

                _staff1 = _staff_per_ss_sl[s1.id]
                _staff2 = _staff_per_ss_sl[s2.id]
                if _staff1.id != _staff2.id:
                    raise ValueError('Really weird notehead overlapping two stafflines'
                                     ' from two different staves: {0}'.format(node.id))

                _staffspace_idx = _ss_sl_idx_wrt_staff[s2.id]
                staff = _staff_and_idx2ss[_staff2.id][_staffspace_idx]
                link_nodes(node, staff)
                # And link to staff:
                _c_staff = _staff_per_ss_sl[staff.id]
                link_nodes(node, _c_staff)

            elif len(overlapped_stafflines) > 2:
                raise ValueError('Really weird notehead overlapping more than 2 stafflines:'
                                 ' {0}'.format(node.id))

    return nodes


###############################################################################


def build_argument_parser():
    parser = argparse.ArgumentParser(description=__doc__, add_help=True,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-a', '--annot', action='store', required=True,
                        help='The annotation file for which the staffline and staff'
                             ' Node relationships should be added.')
    parser.add_argument('-e', '--export', action='store',
                        help='A filename to which the output Nodes'
                             ' should be saved. If not given, will print to'
                             ' stdout.')

    parser.add_argument('-t', '--notehead_staffspace_threshold', action='store', type=float,
                        default=0.2,
                        help='If the ratio of the smaller to the larger lobe w.r.t.'
                             ' an overlapped staffline is lower than this, we consider'
                             ' the notehead to belong to the adjacent staffspace.')

    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Turn on INFO messages.')
    parser.add_argument('--debug', action='store_true',
                        help='Turn on DEBUG messages.')

    return parser


def main(args):
    logging.info('Starting main...')
    _start_time = time.clock()

    ##########################################################################
    logging.info('Import the Node list')
    if not os.path.isfile(args.annot):
        raise ValueError('Annotation file {0} not found!'
                         ''.format(args.annot))
    nodes = read_nodes_from_file(args.annot)

    output_nodes = add_staff_relationships(
        nodes,
        notehead_staffspace_threshold=args.notehead_staffspace_threshold)

    ##########################################################################
    logging.info('Export the combined list.')
    nodes_string = export_node_list(output_nodes)

    if args.export is not None:
        with open(args.export, 'w') as hdl:
            hdl.write(nodes_string)
    else:
        print(nodes_string)

    _end_time = time.clock()
    logging.info('add_staff_reationships.py done in {0:.3f} s'
                 ''.format(_end_time - _start_time))


if __name__ == '__main__':
    parser = build_argument_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
    if args.debug:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

    main(args)



import unittest

from mung.io import parse_node_classes
from mung.node_class import NodeClass


class NodeClassTest(unittest.TestCase):
    def test_note_class_to_string(self):
        # Arrange
        expected = """<NodeClass>
    <Id>1</Id>
    <Name>noteheadFull</Name>
    <GroupName>primitives</GroupName>
    <Color>#FF6689</Color>
</NodeClass>"""
        node_class = NodeClass(1, "noteheadFull", "primitives", "#FF6689")

        # Act
        actual = str(node_class)

        # Assert
        self.assertEqual(actual, expected)

    def test_node_class_parsing(self):
        # Arrange
        expected_number_of_classes = 163

        # Act
        node_classes = parse_node_classes("test/test_data/mff-muscima-classes-annot.xml")

        # Assert
        self.assertEqual(len(node_classes), expected_number_of_classes)
        self.assertEqual(node_classes[0].name, "noteheadFull")


if __name__ == '__main__':
    unittest.main()



"""

**NOTE: This file should become obsolete. The list of classes will be
implemented as a sub-JSON from SMuFL.**

This module implements the :class:`NodeClass`, which
represents one possible :class:`Node` class, such as
a notehead or a time signature. Aside from defining the "vocabulary"
of available object classes for annotation, it also contains
some information about how objects of the given class should
be displayed in the MUSCIMarker annotation software (ordering
related object classes together in menus, implementing a sensible
color scheme, etc.). There is nothing interesting about this class,
we pulled it into the ``mung`` package because the object
grammar (i.e. which relationships are allowed and which are not)
depends on having NodeClass object as its "vocabulary",
and you will probably want to manipulate the data somehow based
on the objects' relationships (like reassembling notes from notation
primitives: notehead plus stem plus flags...), and the grammar
file is a reference for doing that.

NodeClass is a plain old data class, nothing interesting
about it. The only catch is that colors for rendering
in MUSCIMarker are kept as a ``#RRGGBB`` string in the XML
file, but represented in the ``NodeClass.color`` attribute
as a triplet of floats between 0 (``00``) and 255 (``ff``).


The ``___str__()`` method of the class will output the correct
XML representation.

**XML example**

This is what a single NodeClass element might look like::

    <NodeClass>
        <Id>1</Id>
        <Name>notehead-empty</Name>
        <GroupName>note-primitive/notehead-empty</GroupName>
        <Color>#FF7566</Color>
    </NodeClass>

See e.g. ``test/test_data/mff-muscima-classes-annot.xml``,
which is incidentally the real NodeClass list used
for annotating MUSCIMA++.

"""
import logging


class NodeClass(object):
    """Information about the annotation class. We're using it
    mostly to get the color of rendered Node.

    NodeClass is a Plain Old Data class, there is no other
    functionality beyond simply existing and writing itself
    out in the appropriate XML format.
    """

    def __init__(self, class_id, name, group_name, color):
        self.class_id = class_id
        self.name = name
        self.group_name = group_name
        # Parse the string into a RGB spec.
        r, g, b = hex2rgb(color)
        logging.debug('NodeClass {0}: color {1}'.format(name, (r, g, b)))
        self.color = (r, g, b)

    def __str__(self):
        lines = []
        lines.append('<NodeClass>')
        lines.append('    <Id>{0}</Id>'.format(self.class_id))
        lines.append('    <Name>{0}</Name>'.format(self.name))
        lines.append('    <GroupName>{0}</GroupName>'.format(self.group_name))
        lines.append('    <Color>{0}</Color>'.format(rgb2hex(self.color)))
        lines.append('</NodeClass>')
        return '\n'.join(lines)


#######################################################################
# Utility functions for name/writer conversions
_hex_tr = {
    '0': 0,
    '1': 1, '2': 2, '3': 3, '4': 4, '5': 5, '6': 6, '7': 7,
    '8': 8, '9': 9,
    'a': 10, 'b': 11, 'c': 12, 'd': 13, 'e': 14, 'f': 15,
    'A': 10, 'B': 11, 'C': 12, 'D': 13, 'E': 14, 'F': 15,
}
_hex_itr = {v: k for k, v in list(_hex_tr.items())}


def parse_hex(hstr):
    """Convert a hexadecimal number string to integer.

    >>> parse_hex('33')
    51
    >>> parse_hex('abe8')
    44008

    """
    out = 0
    for i, l in enumerate(reversed(hstr)):
        out += (16 ** i) * _hex_tr[l]
    return out


def hex2rgb(hstr):
    """Parse a hex-coded color like '#AA0202' into a floating-point representation.

    >>> hex2rgb('#abe822')
    (0.6705882352941176, 0.9098039215686274, 0.13333333333333333)

    """
    if hstr.startswith('#'):
        hstr = hstr[1:]
    rs, gs, bs = hstr[:2], hstr[2:4], hstr[4:]
    r, g, b = parse_hex(rs), parse_hex(gs), parse_hex(bs)
    return r / 255.0, g / 255.0, b / 255.0


def rgb2hex(rgb):
    """Convert a floating-point representation of R, G, B values
    between 0 and 1 (inclusive) to a hex string (strating with a
    hashmark). Will use uppercase letters for 10 - 15.

    >>> rgb = (0.6705882352941176, 0.9098039215686274, 0.13333333333333333)
    >>> rgb2hex(rgb)
    '#ABE822'

    """
    rgb_int = [int(ch * 255) for ch in rgb]
    return '#' + ''.join(['{:02X}'.format(ch) for ch in rgb_int])



"""This module implements constants that are used inside the pitch,
duration and onset inference algorithm."""
import operator


class InferenceEngineConstants(object):
    """This class stores the constants used for pitch inference."""

    ON_STAFFLINE_RATIO_THRESHOLD = 0.2
    '''Magic number for determining whether a notehead is *on* a leger
    line, or *next* to a leger line: if the ratio between the smaller
    and larger vertical difference of (top, bottom) vs. l.l. (top, bottom)
    is smaller than this, it means the notehead is most probably *NOT*
    on the l.l. and is next to it.'''

    STAFFLINE_CLASS_NAME = 'staffLine'
    STAFFSPACE_CLASS_NAME = 'staffSpace'
    STAFF_CLASS_NAME = 'staff'
    LEGER_LINE_CLASS_NAME = 'legerLine'

    STAFF_CLASS_NAMES = [STAFFLINE_CLASS_NAME, STAFFSPACE_CLASS_NAME, STAFF_CLASS_NAME]
    STAFFLINE_CLASS_NAMES = [STAFFLINE_CLASS_NAME, STAFFSPACE_CLASS_NAME]

    STAFFLINE_LIKE_CLASS_NAMES = [STAFFLINE_CLASS_NAME, LEGER_LINE_CLASS_NAME]
    TIE_CLASS_NAME = 'tie'

    G_CLEF = 'gClef'
    C_CLEF = 'cClef'
    F_CLEF = 'fClef'

    STAFF_RELATED_CLASS_NAMES = {
        'staffGrouping',
        'measureSeparator',
        'keySignature',
        'timeSignature',
        G_CLEF,
        C_CLEF,
        F_CLEF
    }

    SYSTEM_LEVEL_CLASS_NAMES = {
        'staff_grouping',
        'measure_separator',
    }

    NOTEHEAD_CLASS_NAMES = {
        'noteheadFull',
        'noteheadHalf',
        'noteheadWhole',
        'noteheadFullSmall',
        'noteheadHalfSmall',
    }

    NONGRACE_NOTEHEAD_CLASS_NAMES = {
        'noteheadFull',
        'noteheadHalf',
        'noteheadWhole',
    }

    CLEF_CLASS_NAMES = {
        G_CLEF,
        C_CLEF,
        F_CLEF
    }

    KEY_SIGNATURE_CLASS_NAMES = {
        'keySignature',
    }

    MEASURE_SEPARATOR_CLASS_NAMES = {
        'measureSeparator',
    }

    FLAGS_CLASS_NAMES = {
        'flag8thUp',
        'flag8thDown',
        'flag16thUp',
        'flag16thDown',
        'flag32ndUp',
        'flag32ndDown',
        'flag64thUp',
        'flag64thDown',
    }

    BEAM_CLASS_NAMES = {
        'beam',
    }

    FLAGS_AND_BEAMS = set(list(FLAGS_CLASS_NAMES) + list(BEAM_CLASS_NAMES))

    ACCIDENTAL_CLASS_NAMES = {
        'accidentalSharp': 1,
        'accidentalFlat': -1,
        'accidentalNatural': 0,
        'accidentalDoubleSharp': 2,
        'accidentalDoubleFlat': -2,
    }

    MIDI_CODE_RESIDUES_FOR_PITCH_STEPS = {
        0: 'C',
        1: 'C#',
        2: 'D',
        3: 'Eb',
        4: 'E',
        5: 'F',
        6: 'F#',
        7: 'G',
        8: 'Ab',
        9: 'A',
        10: 'Bb',
        11: 'B',
    }
    '''Simplified pitch naming.'''

    # The individual MIDI codes for for the unmodified steps.
    _fs = list(range(5, 114, 12))
    _cs = list(range(0, 121, 12))
    _gs = list(range(7, 116, 12))
    _ds = list(range(2, 110, 12))
    _as = list(range(9, 118, 12))
    _es = list(range(4, 112, 12))
    _bs = list(range(11, 120, 12))

    KEY_TABLE_SHARPS = {
        0: {},
        1: {i: 1 for i in _fs},
        2: {i: 1 for i in _fs + _cs},
        3: {i: 1 for i in _fs + _cs + _gs},
        4: {i: 1 for i in _fs + _cs + _gs + _ds},
        5: {i: 1 for i in _fs + _cs + _gs + _ds + _as},
        6: {i: 1 for i in _fs + _cs + _gs + _ds + _as + _es},
        7: {i: 1 for i in _fs + _cs + _gs + _ds + _as + _es + _bs},
    }

    KEY_TABLE_FLATS = {
        0: {},
        1: {i: -1 for i in _bs},
        2: {i: -1 for i in _bs + _es},
        3: {i: -1 for i in _bs + _es + _as},
        4: {i: -1 for i in _bs + _es + _as + _ds},
        5: {i: -1 for i in _bs + _es + _as + _ds + _gs},
        6: {i: -1 for i in _bs + _es + _as + _ds + _gs + _cs},
        7: {i: -1 for i in _bs + _es + _as + _ds + _gs + _cs + _fs},
    }

    # FROM clef --> TO clef. Imagine this on inline accidental delta
    CLEF_CHANGE_DELTA = {
        G_CLEF: {
            G_CLEF: 0,
            C_CLEF: 6,
            F_CLEF: 12,
        },
        C_CLEF: {
            G_CLEF: -6,
            C_CLEF: 0,
            F_CLEF: 6,
        },
        F_CLEF: {
            G_CLEF: -12,
            C_CLEF: -6,
            F_CLEF: 0,
        }
    }

    PITCH_STEPS = ['C', 'D', 'E', 'F', 'G', 'A', 'B',
                   'C', 'D', 'E', 'F', 'G', 'A', 'B']
    # Wrap around twice for easier indexing.

    ACCIDENTAL_CODES = {'accidentalSharp': '#', 'accidentalFlat': 'b',
                        'accidentalDoubleSharp': 'x', 'accidentalDoubleFlat': 'bb'}

    REST_CLASS_NAMES = {
        'restWhole',
        'restHalf',
        'restQuarter',
        'rest8th',
        'rest16th',
        'rest32nd',
        'rest64th',
        'multiMeasureRest',
    }

    MEAUSURE_LASTING_CLASS_NAMES = {
        'restWhole',
        'multiMeasureRest',
        'repeat1Bar',
    }

    TIME_SIGNATURES = {
        'timeSignature',
    }

    TIME_SIGNATURE_MEMBERS = {
        'timeSigCommon',
        'timeSigCutCommon',
        'numeral0',
        'numeral1',
        'numeral2',
        'numeral3',
        'numeral4',
        'numeral5',
        'numeral6',
        'numeral7',
        'numeral8',
        'numeral9',
    }

    NUMERALS = {
        'numeral0',
        'numeral1',
        'numeral2',
        'numeral3',
        'numeral4',
        'numeral5',
        'numeral6',
        'numeral7',
        'numeral8',
        'numeral9',
    }

    @property
    def classes_affecting_onsets(self):
        """Returns a list of Node class names for objects
        that affect onsets. Assumes notehead and rest durations
        have already been given."""
        output = set()
        output.update(self.NONGRACE_NOTEHEAD_CLASS_NAMES)
        output.update(self.REST_CLASS_NAMES)
        output.update(self.MEASURE_SEPARATOR_CLASS_NAMES)
        output.update(self.TIME_SIGNATURES)
        output.add('repeat1Bar')
        return output

    @property
    def classes_bearing_duration(self):
        """Returns the list of classes that actually bear duration,
        i.e. contribute to onsets of their descendants in the precedence
        graph."""
        output = set()
        output.update(self.NONGRACE_NOTEHEAD_CLASS_NAMES)
        output.update(self.REST_CLASS_NAMES)
        return output




#!/usr/bin/env python
"""This is a script that takes the full-grown notation graph
and recovers for each notehead the pitch to which it corresponds.

Assumptions
-----------

* Clefs are used in a standard way: G-clef on 4th, C-clef on 3rd, F-clef
  on 2nd staffline.
* Key signatures are used in a standard way, so that we can rely on counting
  the accidentals.
* Accidentals are valid up until the end of the bar.

We are currently NOT processing any transpositions.

Representation
--------------

Notes are not noteheads. Pitch is associated with a note, and it is derived
from the notehead's subgraph. The current goal of this exercise is obtaining
MIDI, so we discard in effect information about what is e.g. a G-sharp
and A-flat.

"""
import argparse
import logging
import os
import time

from mung2midi.inference import PitchInferenceEngine, OnsetsInferenceEngine, MIDIBuilder
from mung.io import read_nodes_from_file, export_node_list


def build_argument_parser():
    parser = argparse.ArgumentParser(description=__doc__, add_help=True,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-a', '--annot', action='store', required=True,
                        help='The annotation file for which the staffline and staff'
                             ' Node relationships should be added.')
    parser.add_argument('-e', '--export', action='store',
                        help='A filename to which the output NodeList'
                             ' should be saved. If not given, will print to'
                             ' stdout.')
    parser.add_argument('-m', '--midi', action='store',
                        help='A filename to which to export the MIDI file'
                             ' for the given score.')

    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Turn on INFO messages.')
    parser.add_argument('--debug', action='store_true',
                        help='Turn on DEBUG messages.')
    return parser


##############################################################################


def main(args):
    logging.info('Starting main...')
    _start_time = time.clock()

    # Your code goes here
    if not os.path.isfile(args.annot):
        raise ValueError('Annotation file {0} not found!'
                         ''.format(args.annot))
    nodes = read_nodes_from_file(args.annot)

    pitch_inference_engine = PitchInferenceEngine()
    time_inference_engine = OnsetsInferenceEngine(nodes=nodes)

    logging.info('Running pitch inference.')
    pitches, pitch_names = pitch_inference_engine.infer_pitches(nodes, with_names=True)

    # Export
    logging.info('Adding pitch information to <Data> attributes.')
    for node in nodes:
        if node.id in pitches:
            midi_pitch_code = pitches[node.id]
            pitch_step, pitch_octave = pitch_names[node.id]
            if node.data is None:
                node.data = dict()
            node.data['midi_pitch_code'] = midi_pitch_code
            node.data['normalized_pitch_step'] = pitch_step
            node.data['pitch_octave'] = pitch_octave

    logging.info('Adding duration info to <Data> attributes.')
    durations = time_inference_engine.durations(nodes)
    logging.info('Total durations: {0}'.format(len(durations)))
    for node in nodes:
        if node.id in durations:
            node.data['duration_beats'] = durations[node.id]

    logging.info('Some durations: {0}'.format(sorted(durations.items())[:10]))

    logging.info('Adding onset info to <Data> attributes.')
    onsets = time_inference_engine.onsets(nodes)
    logging.info('Total onsets: {0}'.format(len(onsets)))
    for node in nodes:
        if node.id in onsets:
            node.data['onset_beats'] = onsets[node.id]

    if args.export is not None:
        with open(args.export, 'w') as hdl:
            hdl.write(export_node_list(nodes))
            hdl.write('\n')
    else:
        print(export_node_list(nodes))

    if args.midi is not None:
        midi_builder = MIDIBuilder()
        mf = midi_builder.build_midi(pitches, durations, onsets)
        with open(args.midi, 'wb') as hdl:
            mf.writeFile(hdl)

    _end_time = time.clock()
    logging.info('infer_pitches.py done in {0:.3f} s'.format(_end_time - _start_time))


if __name__ == '__main__':
    parser = build_argument_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
    if args.debug:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

    main(args)



#!/usr/bin/env python
"""``analyze_agreement.py`` is a script that analyzes the agreement between two
annotations of the same file. The script measures:

* Object counts: are they the same?
* Object assignment: given the least-squares mapping of objects
  onto each other, to what extent do they differ?

For an overview of command-line options, call::

  analyze_agreement.py -h

Alignment algorithm
-------------------

The script uses a greedy alignment procedure.

First, it computes for each ``(truth, prediction)`` symbol pair
their recall, precision, and f-score over pixels that fall within
the mask (bounding box overlap may be misleading, mainly for
parallel beams).

Each predicted symbol is then aligned to the ground truth symbol
with the highest f-score. If the symbol classes of a ``(truth, prediction)``
pair do not match, their score gets set to 0. (This can be turned
off using the ``--no_strict_class_names`` option.)

Next, the alignment is cleaned up: if multiple predictions are
aligned to a single ground truth, the one with the highest f-score
is chosen and the other predicted symbols are considered
unaligned.

Computing the output f-score
----------------------------

Finally, we sum all the f-scores of ``(truth, prediction)``
symbol pairs in the alignment.

Ground truth symbols that are not aligned to any predicted object
also contribute a zero to the overall f-score.

"""
import argparse
import collections
import logging
import pprint
import time

import numpy
from typing import List, Tuple, Optional

from mung.io import read_nodes_from_file
from mung.node import Node


def bounding_box_intersection(origin: Tuple[int, int, int, int], intersect: Tuple[int, int, int, int]) \
        -> Optional[Tuple[int, int, int, int]]:
    """Returns the coordinates of the origin bounding box that
    are intersected by the intersect bounding box.

    >>> bounding_box = 10, 100, 30, 110
    >>> other_bbox = 20, 100, 40, 105
    >>> bounding_box_intersection(bounding_box, other_bbox)
    (10, 0, 20, 5)
    >>> bounding_box_intersection(other_bbox, bounding_box)
    (0, 0, 10, 5)
    >>> containing_bbox = 4, 55, 44, 115
    >>> bounding_box_intersection(bounding_box, containing_bbox)
    (0, 0, 20, 10)
    >>> contained_bbox = 12, 102, 22, 108
    >>> bounding_box_intersection(bounding_box, contained_bbox)
    (2, 2, 12, 8)
    >>> non_overlapping_bbox = 0, 0, 3, 3
    >>> bounding_box_intersection(bounding_box, non_overlapping_bbox) is None
    True

    """
    o_t, o_l, o_b, o_r = origin
    t, l, b, r = intersect

    out_top = max(t, o_t)
    out_left = max(l, o_l)
    out_bottom = min(b, o_b)
    out_right = min(r, o_r)

    if (out_top < out_bottom) and (out_left < out_right):
        return out_top - o_t, \
               out_left - o_l, \
               out_bottom - o_t, \
               out_right - o_l
    else:
        return None


def pixel_metrics(truth: Node, prediction: Node) -> Tuple[float, float, float]:
    """Computes the recall, precision and f-score for the prediction
    Node given the truth Node."""
    recall, precision, fscore = 0, 0, 0

    intersection_truth = bounding_box_intersection(truth.bounding_box,
                                                   prediction.bounding_box)
    if intersection_truth is None:
        logging.debug('No intersection for Nodes: t={0},'
                      ' p={1}'.format(truth.bounding_box,
                                      prediction.bounding_box))
        return recall, precision, fscore

    intersection_pred = bounding_box_intersection(prediction.bounding_box,
                                                  truth.bounding_box)

    logging.debug('Found intersection for Nodes: t={0},'
                  ' p={1}'.format(truth.bounding_box,
                                  prediction.bounding_box))

    tt, tl, tb, tr = intersection_truth
    pt, pl, pb, pr = intersection_pred
    crop_truth = truth.mask[tt:tb, tl:tr]
    crop_pred = prediction.mask[pt:pb, pl:pr]

    # Assumes the mask values are 1...

    n_truth = float(truth.mask.sum())
    n_pred = float(prediction.mask.sum())
    n_common = float((crop_truth * crop_pred).sum())

    # There are no zero-pixel objects, but the overlap may be nonzero
    if n_truth == 0:
        recall = 0.0
        precision = 0.0
    elif n_pred == 0:
        recall = 0.0
        precision = 0.0
    else:
        recall = n_common / n_truth
        precision = n_common / n_pred

    if (recall == 0) or (precision == 0):
        fscore = 0
    else:
        fscore = 2 * recall * precision / (recall + precision)

    return recall, precision, fscore


def compute_recall_precision_fscore(truth: List[Node], prediction: List[Node]) \
        -> Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]:
    """Computes Node pixel-level metrics.

    :param truth: A list of the ground truth Nodes.

    :param prediction: A list of the predicted Nodes.

    :returns: Three matrices with shape ``(len(truth), len(prediction)``:
        recall, precision, and f-score for each truth/prediction Node
        pair. Truth Nodes are rows, prediction columns.
    """
    recall = numpy.zeros((len(truth), len(prediction)))
    precision = numpy.zeros((len(truth), len(prediction)))
    fscore = numpy.zeros((len(truth), len(prediction)))

    for i, t in enumerate(truth):
        for j, p in enumerate(prediction):
            r, p, f = pixel_metrics(t, p)
            recall[i, j] = r
            precision[i, j] = p
            fscore[i, j] = f

    return recall, precision, fscore


def align_nodes(truth: List[Node], prediction: List[Node], fscore=None) -> List[Tuple[int, int]]:
    """Aligns prediction Nodes to truth.

    :param truth: A list of the ground truth Nodes.

    :param prediction: A list of the predicted Nodes.

    :returns: A list of (t, p) pairs of Node indices into
        the truth and prediction lists. There will be one
        pair for each predicted symbol.
    """
    if fscore is None:
        _, _, fscore = compute_recall_precision_fscore(truth, prediction)

    # For each prediction (column), pick the highest-scoring
    # True symbol.
    closest_truths = list(fscore.argmax(axis=0))

    # Checking for duplicate "best ground truth" alignments.
    # This does *not* check for duplicates in the sense
    # "multiple predictions aligned to the same truth"
    # or "multiple truths aligned to the same prediction",
    # it only acts as a tie-breaker in case one prediction overlaps
    # to the same degree multiple truth objects (e.g. a single sharp
    # in a key signature).
    closest_truth_distance = [fscore[ct, j]
                              for j, ct in enumerate(closest_truths)]
    equidistant_closest_truths = [[i for i, x in enumerate(fscore[:, j])
                                   if x == ct]
                                  for j, ct in enumerate(closest_truth_distance)]

    class_name_aware_closest_truths = []
    for j, ects in enumerate(equidistant_closest_truths):
        best_truth_i = int(ects[0])

        # If there is more than one tied best choice,
        # try to choose the truth Node that has the same
        # class as the predicted Node.
        if len(ects) > 1:
            ects_c = {truth[int(i)].class_name: i for i in ects}
            j_class_name = prediction[j].class_name
            if j_class_name in ects_c:
                best_truth_i = int(ects_c[j_class_name])

        class_name_aware_closest_truths.append(best_truth_i)

    alignment = [(t, p) for p, t in enumerate(class_name_aware_closest_truths)]
    return alignment


def compute_recall_precision_fscore_given_an_alignment(alignment: List[Tuple[int, int]], individual_recalls,
                                                       individual_precisions, n_not_aligned: int = 0,
                                                       strict_classnames: bool = True,
                                                       truths: List[Node] = None, predictions: List[Node] = None)\
        -> Tuple[float, float, float]:
    if strict_classnames:
        if not truths:
            raise ValueError('If strict_classnames is requested, must supply truths Nodes!')
        if not predictions:
            raise ValueError('If strict_classnames is requested, must supply predictions Nodes!')

    total_recall, total_precision = 0, 0

    for i, j in alignment:

        # Check for strict_classnames only at this stage.
        # The purpose is: if two people mark the same object
        # differently, we do want to know "it should be aligned
        # to each other, but the classes don't fit" -- we don't
        # want to maybe align it to an overlapping object of
        # the corresponding wrong class.
        if strict_classnames:
            t_c = truths[i]
            p_c = predictions[j]
            if t_c.class_name != p_c.class_name:
                continue

        total_recall += individual_recalls[i, j]
        total_precision += individual_precisions[i, j]

    # This is not correct...? What about the zeros?
    #  - Prediction with no GT is in the alignment, so it counts towards
    #    the len(alignment) denominator.
    #  - GT with no prediction aligned to it contributes zero, but is
    #    not counted towards the denominator.
    total_recall /= len(alignment) + n_not_aligned
    total_precision /= len(alignment) + n_not_aligned

    if (total_recall == 0) or (total_precision == 0):
        total_fscore = 0.0
    else:
        total_fscore = 2 * total_recall * total_precision / (total_recall + total_precision)
    return total_recall, total_precision, total_fscore


##############################################################################


def build_argument_parser():
    parser = argparse.ArgumentParser(description=__doc__, add_help=True,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-t', '--true', action='store', required=True,
                        help='The Nodes file you want to consider'
                             ' ground truth.')
    parser.add_argument('-p', '--prediction', action='store', required=True,
                        help='The Nodes file you want to consider'
                             ' the prediction.')

    parser.add_argument('-e', '--export', action='store',
                        help='If set, will export the problematic Nodes'
                             ' to this file.')

    parser.add_argument('--analyze_alignment', action='store_true',
                        help='If set, will check whether the alignment is 1:1,'
                             ' and print out the irregularities.')
    parser.add_argument('--analyze_classnames', action='store_true',
                        help='If set, will check whether the Nodes aligned'
                             ' to each other have the same class labels'
                             ' and print out the irregularities.')
    parser.add_argument('--no_strict_classnames', action='store_true',
                        help='If set, will not require aligned objects\' classnames'
                             ' to match before computing pixel-wise overlap'
                             ' metrics.')
    parser.add_argument('--log_alignment', action='store_true',
                        help='Print how the true and predicted objects are'
                             ' paired.')

    parser.add_argument('--print_fscore_only', action='store_true',
                        help='If set, only print the total F-score number.'
                             ' Useful for using in an automated pipeline.')

    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Turn on INFO messages.')
    parser.add_argument('--debug', action='store_true',
                        help='Turn on DEBUG messages.')

    return parser


def main(args):
    logging.info('Starting main...')
    _start_time = time.clock()

    # The algorithm:
    #  - build the cost function(s) for a pair of Nodes
    #  - align the objects, using the cost function

    # First alignment: try just matching a predicted object to the nearest
    # true object.
    # First distance function: proportion of shared pixels.
    # Rule: if two objects don't share a pixel, they cannot be considered related.
    # Object classes do not factor into this so far.

    ground_truth_nodes = read_nodes_from_file(args.true)
    predicted_nodes = read_nodes_from_file(args.prediction)

    _parse_time = time.clock()
    logging.info('Parsing {0} true and {1} prediction Nodes took {2:.2f} s'
                 ''.format(len(ground_truth_nodes), len(predicted_nodes), _parse_time - _start_time))

    recall, precision, fscore = compute_recall_precision_fscore(ground_truth_nodes, predicted_nodes)

    _rpf_time = time.clock()
    logging.info('Computing {0} entries of r/p/f matrices took {1:.2f} s'
                 ''.format(len(ground_truth_nodes) * len(predicted_nodes), _rpf_time - _parse_time))

    alignment_tp = align_nodes(ground_truth_nodes, predicted_nodes, fscore=fscore)
    alignment_pt = align_nodes(predicted_nodes, ground_truth_nodes, fscore=fscore.T)

    # Intersect alignments
    _aln_tp_set = frozenset(alignment_tp)
    alignment_tp_symmetric = [(t, p) for p, t in alignment_pt
                              if (t, p) in _aln_tp_set
                              and (ground_truth_nodes[t].class_name == predicted_nodes[p].class_name)]
    truth_not_aligned = [t for p, t in alignment_pt
                         if (t, p) not in alignment_tp_symmetric]
    n_truth_not_aligned = len(truth_not_aligned)
    preds_not_aligned = [p for t, p in alignment_tp
                         if (t, p) not in alignment_tp_symmetric]
    n_preds_not_aligned = len(preds_not_aligned)
    n_not_aligned = n_truth_not_aligned + n_preds_not_aligned

    _aln_time = time.clock()
    logging.info('Computing alignment took {0:.2f} s'
                 ''.format(_aln_time - _rpf_time))

    # Now compute agreement: precision and recall on pixels
    # of the aligned Nodes.

    # We apply strict classnames only here, after the Nodes have been
    # aligned to each other using pixel metrics.
    strict_classnames = (not args.no_strict_classnames)
    total_r, total_p, total_f = compute_recall_precision_fscore_given_an_alignment(alignment_tp_symmetric, recall, precision,
                                                                                   n_not_aligned=n_not_aligned,
                                                                                   strict_classnames=strict_classnames,
                                                                                   truths=ground_truth_nodes,
                                                                                   predictions=predicted_nodes)

    if not args.print_fscore_only:
        print('Truth objs.:\t{0}'.format(len(ground_truth_nodes)))
        print('Pred. objs.:\t{0}'.format(len(predicted_nodes)))
        print('Aligned objs.:\t{0}'.format(len(alignment_tp_symmetric)))
        print('==============================================')
        print('Recall:\t\t{0:.3f}\nPrecision:\t{1:.3f}\nF-score:\t{2:.3f}'
              ''.format(total_r, total_p, total_f))
        print('')
    else:
        print('{0:.3f}'.format(total_f))
        return

    if args.log_alignment:
        print('==============================================')
        print('Alignments:\n{0}'.format('\n'.join([
            '({0}: {1}) -- ({2}: {3})'.format(ground_truth_nodes[t].id, ground_truth_nodes[t].class_name,
                                              predicted_nodes[p].id, predicted_nodes[p].class_name)
            for t, p in alignment_tp_symmetric
        ])))
        print('Truth, not aligned:\n{0}'.format('\n'.join(['({0}: {1})'.format(ground_truth_nodes[t].id, ground_truth_nodes[t].class_name)
                                                           for t in truth_not_aligned])))
        print(
            'Preds, not aligned:\n{0}'.format('\n'.join(['({0}: {1})'.format(predicted_nodes[p].id, predicted_nodes[p].class_name)
                                                         for p in preds_not_aligned])))

    ##########################################################################
    # Check if the alignment is a pairing -- find truth objects
    # with more than one prediction aligned to them.
    if args.analyze_alignment:
        t_aln_dict = collections.defaultdict(list)
        for i, j in alignment_tp_symmetric:
            t_aln_dict[i].append(predicted_nodes[j])

        multiple_truths = [ground_truth_nodes[i] for i in t_aln_dict
                           if len(t_aln_dict[i]) > 1]
        multiple_truths_aln_dict = {t: t_aln_dict[t]
                                    for t in t_aln_dict
                                    if len(t_aln_dict[t]) > 1}

        print('Truth multi-aligned Node classes:\n{0}'
              ''.format(pprint.pformat(
            {(ground_truth_nodes[t].id, ground_truth_nodes[t].class_name): [(p.id, p.class_name)
                                                  for p in t_aln_dict[t]]
             for t in multiple_truths_aln_dict})))

    ##########################################################################
    # Check if the aligned objects have the same classes
    if args.analyze_classnames:
        different_classnames_pairs = []
        for i, j in alignment_tp_symmetric:
            if ground_truth_nodes[i].class_name != predicted_nodes[j].class_name:
                different_classnames_pairs.append((ground_truth_nodes[i], predicted_nodes[j]))
        print('Aligned pairs with different class_names:\n{0}'
              ''.format('\n'.join(['{0}.{1}\t{2}.{3}'
                                   ''.format(t.id, t.class_name, p.id, p.class_name)
                                   for t, p in different_classnames_pairs])))

    _end_time = time.clock()
    logging.info('analyze_agreement.py done in {0:.3f} s'.format(_end_time - _start_time))


if __name__ == '__main__':
    parser = build_argument_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
    if args.debug:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

    main(args)



#!/usr/bin/env python
"""``get_images_from_muscima.py`` is a script that copies out the images
for which MUSCIMA++ provides symbol annotations from a download
of the CVC-MUSCIMA staff removal dataset.

You have to download this dataset first and provide a path to its
root directory (meaning the directory which contains subdirs for
the individual CVC-MUSCIMA distortions)
to this script. Either supply it directly using the ``-r`` option,
or set a ``CVC_MUSCIMA_ROOT`` environmental variable.

Example invocation::

    get_images_from_muscima.py -o ./images -i 4:10 17:8 5:12 21:10 34:3

MUSCIMA++ 0.9 provides a file with the writer:number pairs for its 140
annotated images in this format, which you can feed to the script
with::

    get_images_from_muscima.py [...] -i `cat path/to/MUSCIMA++/specifications/cvc-muscima-image-list.txt

For an overview of all command-line options, call::

  get_images_from_muscima.py -h

"""
import argparse
import logging
import os
import time

import shutil

from mung.dataset import CvcMuscimaDataset, CVC_MUSCIMA_ROOT


def build_argument_parser():
    parser = argparse.ArgumentParser(description=__doc__, add_help=True,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-r', '--root', action='store',
                        default=CVC_MUSCIMA_ROOT,
                        help='CVC-MUSCIMA dataset root directory (should'
                             ' contain subdirectories named after the'
                             ' CVC-MUSCIMA distortions).')
    parser.add_argument('-o', '--outdir', action='store',
                        help='Output directory for the copied files.'
                             ' If it does not exist, it will be created.')
    parser.add_argument('-i', '--items', action='store', nargs='+',
                        help='A list of writer:page pairs, such as 22:4.')
    parser.add_argument('-f', '--format', action='store',
                        default='CVC-MUSCIMA_W-{w:02}_N-{n:02}_D-ideal',
                        help='The desired output filenames. {w} and {n}'
                             ' stand for writer and page number: for'
                             ' item 4:22, for instance, the filename'
                             ' would be CVC_MUSCIMA_W-22_N-04_D-ideal.png'
                             ' (the *.png suffix is retained from the'
                             ' corresponding CVC-MUSCIMA file).')
    parser.add_argument('-m', '--mode', action='store', default='symbol',
                        help='The CVC-MUSCIMA image mode: \'full\', \'symbol\','
                             ' or \'staff_only\'.')

    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Turn on INFO messages.')
    parser.add_argument('--debug', action='store_true',
                        help='Turn on DEBUG messages.')

    return parser


##############################################################################


def main(args):
    logging.info('Starting main...')
    _start_time = time.clock()

    if not os.path.isdir(args.outdir):
        os.mkdir(args.outdir)

    dataset = CvcMuscimaDataset(root=args.root)

    writers = []
    pages = []
    for item in args.items:
        ws, ps = item.split(':')
        writers.append(int(ws))
        pages.append(int(ps))

    for w, p in zip(writers, pages):
        imfile = dataset.imfile(page=p, writer=w,
                                distortion='ideal',
                                mode=args.mode)
        # Format the filename
        _, out_ext = os.path.splitext(imfile)
        out_fname = args.format.format(w=w, n=p) + out_ext
        out_file = os.path.join(args.outdir, out_fname)
        # Copy the file
        shutil.copyfile(imfile, out_file)

    _end_time = time.clock()
    logging.info('get_images_from_muscima.py done in {0:.3f} s'
                 ''.format(_end_time - _start_time))


if __name__ == '__main__':
    parser = build_argument_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
    if args.debug:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

    main(args)



"""``analyze_tracking_log.py`` is a script that performs a quick and dirty analysis
of a MUSCIMarker event log. It is not necessary for using the dataset,
but you might want it running if you are annotating something with MUSCIMarker.

For an overview of command-line options, call::

  analyze_tracking_log.py -h

What does the script track?

* Number of minutes/hours worked
* Speed: how much was done in total?
* Densities: frequency of events (calls) per minute/hour

Visualizations:

* Timing visualization

Also, convert to CSV, to make it grep-able? First: fixed-name cols,
then: args dict, formatted as key=value,key=value

"""
import argparse
import codecs
import collections
import itertools
import json
import logging
import operator
import os
import pprint
import time

import matplotlib.pyplot as plt
import numpy
from typing import Dict, Tuple

from mung.io import read_nodes_from_file


def freqdict(l, sort=True) -> Dict:
    out = collections.defaultdict(int)
    for item in l:
        out[item] += 1
    if sort:
        s_out = collections.OrderedDict()
        for k, v in sorted(list(out.items()), key=operator.itemgetter(1), reverse=True):
            s_out[k] = v
        out = s_out
    return out


##############################################################################

def is_annotation_package(path):
    """Checks that the given path is an annotation package."""
    if not os.path.isdir(path):
        return False
    subdirs = os.listdir(path)
    if 'source_images' not in subdirs:
        return False
    if 'annotations' not in subdirs:
        return False
    if 'annotation_logs' not in subdirs:
        return False
    return True


def logs_from_package(package):
    """Collects all log file names (with complete paths) from the given package.

    :param package: Path to the annotations package.

    :return: List of filenames (full paths).
    """
    logging.info('Collecting log files from package {0}'.format(package))
    if not os.path.isdir(package):
        raise OSError('Package {0} not found!'.format(package))
    log_path = os.path.join(package, 'annotation_logs')
    if not os.path.isdir(log_path):
        raise ValueError('Package {0}: annotation_logs not found, probably not a package.'
                         ''.format(package))
    # Collect all log days
    log_days = os.listdir(log_path)

    # Dealing with people who copied the entire .muscimarker-tracking directory
    # (potentially without the dot, as just "muscimarker-tracking")
    if len(log_days) == 0:
        logging.info('No logs in package {0}!'.format(package))
        return []

    if log_days[-1].endswith('muscimarker-tracking'):
        log_path = os.path.join(log_path, log_days[-1])
        log_days = os.listdir(log_path)

    log_files = []
    for day in log_days:

        # .DS_store and other hidden files
        if day.startswith('.'):
            continue

        # Dealing with people who copied only the JSON files
        if day.endswith('json'):
            logging.info('Found log file that is not inside a day dir: {0}'
                         ''.format(day))
            log_files.append(os.path.join(log_path, day))
            continue

        if day.endswith('xml'):
            logging.info('Log file is for some reason XML instead of JSON; copied wrong files???')
            continue

        day_log_path = os.path.join(log_path, day)
        day_log_files = [os.path.join(day_log_path, l)
                         for l in os.listdir(day_log_path)]
        log_files += day_log_files
    logging.info('In package {0}: found {1} log files.'
                 ''.format(package, len(log_files)))
    logging.debug('In package {0}: log files:\n{1}'
                  ''.format(package, pprint.pformat(log_files)))
    return log_files


def try_correct_crashed_json(fname):
    """Attempts to correct an incomplete JSON list file: if MUSCIMarker
    crashed, the items list would not get correctly closed. We attempt
    to remove the last comma and add a closing bracket (`]`) on a new
    line instead, and return the object as a (unicode) string.

    >>> json = '''
    ... [
    ...   {'something': 'this', 'something': 'that'},'''

    """
    with open(fname, 'r') as hdl:
        lines = [l.rstrip() for l in hdl]
    if lines[-1][-1] == ',':
        logging.info('Correcting JSON: found hanging comma!')
        lines[-1] = lines[-1][:-1]
        lines.append(']')
        return '\n'.join(lines)

    else:
        logging.info('No hanging comma, cannot deal with this situation.')
        return None


def unique_logs(event_logs):
    """Checks that the event logs are unique using the start event
    timestamp. Returns a list of unique event logs. If two have the same
    timestamp, the first one is used.

    For logging purposes, expects a dict of event logs. Keys are log file names,
    values are the event lists.
    """
    unique = collections.OrderedDict()
    for log_file, l in event_logs.items():
        if len(l) < 1:
            logging.info('Got an empty log from file {0}'.format(log_file))
            continue
        init_event = l[0]
        if '-time-' not in init_event:
            raise ValueError('Got a non-event log JSON list, file {0}! Supposed init event: {1}'
                             ''.format(log_file, init_event))
        init_time = init_event['-time-']
        if init_time in unique:
            logging.info('Found non-unique event log {0} with timestamp {1} ({2} events)!'
                         ' Using first ({3} events).'
                         ''.format(log_file, init_time, len(l), len(unique[init_time])))
        else:
            unique[init_time] = l
    return list(unique.values())


##############################################################################
# Counting results


def annotations_from_package(package: str):
    """Collect all annotation XML files (with complete paths)
    from the given package."""
    logging.info('Collecting annotation files from package {0}'.format(package))
    if not os.path.isdir(package):
        raise OSError('Package {0} not found!'.format(package))
    annot_path = os.path.join(package, 'annotations')
    if not os.path.isdir(annot_path):
        raise ValueError('Package {0}: annotations not found, probably not a package.'
                         ''.format(package))

    # Collect all annotations
    annotation_files = [os.path.join(annot_path, f)
                        for f in os.listdir(annot_path) if f.endswith('.xml')]
    return annotation_files


def count_nodes(annot_file):
    return len(read_nodes_from_file(annot_file))


def count_nodes_and_relationships(annot_file: str) -> Tuple[int, int]:
    nodes = read_nodes_from_file(annot_file)
    n_inlinks = 0
    for node in nodes:
        if node.inlinks is not None:
            n_inlinks += len(node.inlinks)
    return len(nodes), n_inlinks


##############################################################################
# Visualization

def events_by_time_units(events, seconds_per_unit=60):
    """Puts the events into bins that correspond to equally spaced
    intervals of time. The length of time covered by one bin is
    given by seconds_per_unit."""
    # Get first event time
    start_time = min([float(e['-time-']) for e in events])

    # The events do not have to come in-order
    bins = collections.defaultdict(list)
    for e in events:
        t = float(e['-time-'])
        n_bin = int(t - start_time) // int(seconds_per_unit)
        bins[n_bin].append(e)

    return bins


def plot_events_by_time(events, type_key='-fn-'):
    """Simple scatterplot visualization.

    All events are expected to have a -fn- component."""
    fns = [e['-fn-'] for e in events]
    # Assign numbers to tracked fns
    fns_by_freq = {f: len([e for e in fns if e == f]) for f in set(fns)}
    fn_dict = {f: i for i, f in enumerate(sorted(list(fns_by_freq.keys()),
                                                 reverse=True,
                                                 key=lambda k: fns_by_freq[k]))}

    min_time = float(events[0]['-time-'])

    dataset = numpy.zeros((len(events), 2))
    for i, e in enumerate(events):
        dataset[i][0] = float(e['-time-']) - min_time
        dataset[i][1] = fn_dict[e[type_key]]

    # Now visualize
    plt.scatter(dataset[:, 0], dataset[:, 1])


def format_as_timeflow_csv(events, delimiter='\t'):
    """There is a cool offline visualization tool caled TimeFlow,
    which has a timeline app. It needs a pretty specific CSV format
    to work, though."""
    # What we need:
    #  - ID
    #  - Date (human?)
    #  - The common fields:
    min_second = int(min([float(e['-time-']) for e in events]))

    def format_date(e):
        # return '-'.join(reversed(time_human.replace(':', '-').split('__')))
        # time_human = e['-time-human-']
        time = float(e['-time-'])
        return str(int(time) - min_second)

    # Collect all events that are in the data.
    event_fields = freqdict(list(itertools.chain(*[list(e.keys()) for e in events])))
    output_fields = ['ID', 'Date'] + list(event_fields.keys())
    n_fields = len(output_fields)

    field2idx = {f: i + 2 for i, f in enumerate(event_fields.keys())}
    event_table = [['' for _ in range(n_fields)] for _ in events]
    for i, e in enumerate(events):
        event_table[i][0] = str(i)
        event_table[i][1] = format_date(e)  # format_date(e['-time-human-'])
        for k, v in e.items():
            event_table[i][field2idx[k]] = v

    # Add labels to event table to get the complete data
    # that should be formatted as TSV
    output_data = [output_fields] + event_table
    output_lines = ['\t'.join(row) for row in output_data]
    output_string = '\n'.join(output_lines)
    return output_string


##############################################################################


def build_argument_parser():
    parser = argparse.ArgumentParser(description=__doc__, add_help=True,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-i', '--inputs', nargs='+', action='store',
                        help='Log files to be analyzed.')
    parser.add_argument('-p', '--packages', nargs='+', action='store',
                        help='Annotation package. If set, will pull'
                             ' all log files in the package.')
    parser.add_argument('-a', '--annotator', action='store',
                        help='Annotator. If set, will pull all log files'
                             ' from all packages in the given person\'s'
                             ' annotation directory')

    parser.add_argument('-c', '--count_annotations', action='store_true',
                        help='If given, will collect annotation files from the'
                             ' supplied packages (or per-annotator packages)'
                             ' and compute object/rel counts and efficiency statistics.')
    parser.add_argument('--no_training', action='store_true',
                        help='If given, will ignore packages with "training" in their name.')

    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Turn on INFO messages.')
    parser.add_argument('--debug', action='store_true',
                        help='Turn on DEBUG messages.')

    return parser


def main(args):
    logging.info('Starting main...')
    _start_time = time.clock()

    if args.annotator is not None:
        logging.info('Collecting annotation packages for annotator {0}'
                     ''.format(args.annotator))
        # Collect all packages, incl. training
        packages = []
        for d in os.listdir(args.annotator):
            package_candidate = os.path.join(args.annotator, d)
            if not is_annotation_package(package_candidate):
                continue
            packages.append(package_candidate)

        logging.info('Found: {0} packages'.format(len(packages)))

        args.packages = packages

    if args.packages is not None:
        logging.info('Collecting log files for {0} packages.'.format(len(args.packages)))
        log_files = []
        for package in args.packages:
            current_log_files = logs_from_package(package)
            log_files += current_log_files

        logging.info('Found: {0} log files'.format(len(log_files)))
        args.input = log_files

    log_data_per_file = {}
    for input_file in args.input:
        if not os.path.isfile(input_file):
            raise ValueError('Log file {0} not found!'.format(input_file))

        current_log_data = []

        with codecs.open(input_file, 'r', 'utf-8') as hdl:
            try:
                current_log_data = json.load(hdl)

            except ValueError:
                logging.info('Could not parse JSON file {0}'.format(input_file))
                logging.info('Attempting to correct file.')
                corrected = try_correct_crashed_json(input_file)
                if corrected is not None:
                    logging.info('Attempting to parse corrected JSON.')
                    try:
                        current_log_data = json.loads(corrected)
                    except ValueError:
                        logging.warning('Could not even parse corrected JSON, skipping file {0}.'.format(input_file))
                        # raise
                    logging.info('Success!')
                else:
                    logging.info('Unable to correct JSON, skipping file.')

        log_data_per_file[input_file] = current_log_data

    logging.info('Checking logs for uniqueness. Started with {0} log files.'
                 ''.format(len(log_data_per_file)))
    log_data_per_file = unique_logs(log_data_per_file)
    logging.info('After uniqueness check: {0} logs left.'.format(len(log_data_per_file)))

    log_data = [e for e in itertools.chain(*log_data_per_file)]
    if len(log_data) == 0:
        print('Received no log data! Skipping ahead to count annotations.')
        n_minutes = None
        n_hours = None
    else:
        logging.info('Parsed {0} data items.'.format(len(log_data)))
        # Your code goes here
        # raise NotImplementedError()

        # Frequency by -fn-:
        freq_by_fn = freqdict([l.get('-fn-', None) for l in log_data])

        by_minute = events_by_time_units(log_data)
        by_minute_freq = {k: len(v) for k, v in list(by_minute.items())}
        n_minutes = len(by_minute)

        print('# minutes worked: {0}'.format(n_minutes))
        n_hours = n_minutes / 60.0
        print('# hours worked: {0:.2f}'.format(n_hours))
        print('CZK@120: {0:.3f}'.format(n_hours * 120))
        print('CZK@150: {0:.3f}'.format(n_hours * 150))
        print('CZK@180: {0:.3f}'.format(n_hours * 180))
        print('Avg. events per minute: {0}'.format(float(len(log_data)) / n_minutes))

    if args.count_annotations:
        if args.packages is None:
            raise ValueError('Cannot count annotations if no packages are given!')

        number_of_nodes = 0
        number_of_relationships = 0
        for package in args.packages:
            annot_files = annotations_from_package(package)
            for f in annot_files:
                n_c, n_r = count_nodes_and_relationships(f)
                number_of_nodes += n_c
                number_of_relationships += n_r

        print('Total Nodes: {0}'.format(number_of_nodes))
        print('Total Relationships: {0}'.format(number_of_relationships))
        if n_minutes is not None:
            print('Nodes per minute: {0:.2f}'.format(number_of_nodes / float(n_minutes)))

    _end_time = time.clock()
    logging.info('analyze_tracking_log.py done in {0:.3f} s'.format(_end_time - _start_time))


##############################################################################


if __name__ == '__main__':
    parser = build_argument_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
    if args.debug:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

    main(args)



"""This module implements an abstraction over a notation graph, and
functions for manipulating notation graphs."""
import copy
import logging
import operator
from queue import Queue

from typing import List, Union, Dict, Set, Tuple, Iterable, Optional

from mung.node import Node
from mung.constants import InferenceEngineConstants as _CONST


class NotationGraphError(ValueError):
    pass


class NotationGraphUnsupportedError(NotImplementedError):
    pass


class NotationGraph(object):
    """The NotationGraph class is the abstraction for a notation graph."""

    def __init__(self, nodes: List[Node]):
        """Initialize the notation graph with a list of Nodes."""
        self.__nodes = nodes
        self.__id_to_node_mapping = {node.id: node for node in self.__nodes}  # type: Dict[int, Node]

    def __len__(self):
        return len(self.__nodes)

    def __to_id(self, node_or_id: Union[Node, int]) -> int:
        if isinstance(node_or_id, Node):
            return node_or_id.id
        else:
            return node_or_id

    @property
    def edges(self) -> Set[Tuple[int, int]]:
        edges = set()
        for node in self.__nodes:
            for t in node.outlinks:
                if (node.id, t) not in edges:
                    edges.add((node.id, t))
        return edges

    @property
    def vertices(self) -> List[Node]:
        return self.__nodes

    def __getitem__(self, node_id: int) -> Node:
        """Returns a Node based on its id."""
        return self.__id_to_node_mapping[node_id]

    def children(self, node_or_id: Union[Node, int], class_filter: Optional[Iterable[str]] = None) -> List[Node]:
        """Find all children of the given node. class_filter can be used to only get children of a particular class. """
        node_id = self.__to_id(node_or_id)
        if node_id not in self.__id_to_node_mapping:
            raise ValueError('Node {0} not in graph!'.format(self.__id_to_node_mapping[node_id].id))

        parent = self.__id_to_node_mapping[node_id]
        children = []
        for child_id in parent.outlinks:
            if child_id in self.__id_to_node_mapping:
                child = self.__id_to_node_mapping[child_id]
                if class_filter is None:
                    children.append(child)
                elif child.class_name in class_filter:
                    children.append(child)
        return children

    def parents(self, node_or_id: Union[Node, int], class_filter: Optional[Iterable[str]] = None) -> List[Node]:
        """Find all parents of the given node.  class_filter can be used to only get parents of a particular class. """
        node_id = self.__to_id(node_or_id)
        if node_id not in self.__id_to_node_mapping:
            raise ValueError('Node {0} not in graph!'.format(self.__id_to_node_mapping[node_id].id))

        child = self.__id_to_node_mapping[node_id]
        parents = []
        for parent_ids in child.inlinks:
            if parent_ids in self.__id_to_node_mapping:
                parent = self.__id_to_node_mapping[parent_ids]
                if class_filter is None:
                    parents.append(parent)
                elif parent.class_name in class_filter:
                    parents.append(parent)
        return parents

    def descendants(self, node_or_id: Union[Node, int], class_filter: Optional[Iterable[str]] = None) -> List[Node]:
        """Find all descendants of the given node."""
        node_id = self.__to_id(node_or_id)

        descendant_ids = []
        queue = Queue()
        queue.put(node_id)
        while not queue.empty():
            current_node_id = queue.get()
            if current_node_id != node_id:
                descendant_ids.append(current_node_id)
            children = self.children(current_node_id, class_filter=class_filter)
            children_node_ids = [child.id for child in children]
            for child_id in children_node_ids:
                if child_id not in queue:
                    queue.put(child_id)
        return [self.__id_to_node_mapping[o] for o in descendant_ids]

    def ancestors(self, node_or_id: Union[Node, int], class_filter: Optional[Iterable[str]] = None) -> List[Node]:
        """Find all ancestors of the given node."""
        node_id = self.__to_id(node_or_id)

        ancestor_node_ids = []
        queue = Queue()
        queue.put(node_id)
        while not queue.empty():
            current_node_id = queue.get()
            if current_node_id != node_id:
                ancestor_node_ids.append(current_node_id)
            parents = self.parents(current_node_id, class_filter=class_filter)
            parent_node_ids = [parent.id for parent in parents]
            for parent_id in parent_node_ids:
                if parent_id not in queue:
                    queue.put(parent_id)

        return [self.__id_to_node_mapping[objid] for objid in ancestor_node_ids]

    def has_children(self, node_or_id: Union[Node, int], class_filter: Optional[Iterable[str]] = None) -> bool:
        children = self.children(node_or_id, class_filter=class_filter)
        return len(children) > 0

    def has_parents(self, node_or_id: Union[Node, int], class_filter: Optional[Iterable[str]] = None) -> bool:
        parents = self.parents(node_or_id, class_filter=class_filter)
        return len(parents) > 0

    def is_child_of(self, child_node_or_id: Union[Node, int],
                    parent_node_or_id: Union[Node, int]) -> bool:
        """Check whether the first symbol is a child of the second symbol."""
        child_id = self.__to_id(child_node_or_id)
        parent_id = self.__to_id(parent_node_or_id)

        parent = self.__id_to_node_mapping[parent_id]
        if child_id in parent.outlinks:
            return True
        else:
            return False

    def is_parent_of(self, parent_node_or_id: Union[Node, int],
                     child_node_or_id: Union[Node, int]) -> bool:
        """Check whether the first symbol is a parent of the second symbol."""
        return self.is_child_of(child_node_or_id, parent_node_or_id)

    def is_stem_direction_above(self, notehead: Node, stem: Node) -> bool:
        """Determines whether the given stem of the given notehead
        is above it or below. This is not trivial due to chords.
        """
        if notehead.id not in self.__id_to_node_mapping:
            raise NotationGraphError('Asking for notehead which is not in graph: {0}'.format(notehead.id))

        # This works even if there is just one. There should always be one.
        sibling_noteheads = self.parents(stem, class_filter=_CONST.NOTEHEAD_CLASS_NAMES)
        if notehead not in sibling_noteheads:
            raise ValueError('Asked for stem direction, but notehead {0} is'
                             ' unrelated to given stem {1}!'
                             ''.format(notehead.id, stem.id))

        topmost_notehead = min(sibling_noteheads, key=lambda x: x.top)
        bottom_notehead = max(sibling_noteheads, key=lambda x: x.bottom)

        d_top = topmost_notehead.top - stem.top
        d_bottom = stem.bottom - bottom_notehead.bottom

        return d_top > d_bottom

    def is_symbol_above_notehead(self, notehead: Node, other: Node,
                                 compare_on_intersect: bool = False) -> bool:
        """Determines whether the given other symbol is above
        the given notehead.

        This is non-trivial because the other may reach above *and* below
        the given notehead, if it is long and slanted (beam, slur, ...).
        A horizontally intersecting subset of the mask of the other symbol
        is used to determine its vertical bounds relevant to the given object.
        """
        if notehead.right <= other.left:
            # No horizontal overlap, notehead to the left
            beam_submask = other.mask[:, :1]
        elif notehead.left >= other.right:
            # No horizontal overlap, notehead to the right
            beam_submask = other.mask[:, -1:]
        else:
            h_bounds = (max(notehead.left, other.left),
                        min(notehead.right, other.right))

            beam_submask = other.mask[:, (h_bounds[0] - other.left):(h_bounds[1] - other.left)]

        # Get vertical bounds of beam submask
        other_submask_hsum = beam_submask.sum(axis=1)
        other_submask_top = min([i for i in range(beam_submask.shape[0])
                                 if other_submask_hsum[i] != 0]) + other.top
        other_submask_bottom = max([i for i in range(beam_submask.shape[0])
                                    if other_submask_hsum[i] != 0]) + other.top
        if (notehead.top <= other_submask_top <= notehead.bottom) \
                or (other_submask_bottom <= notehead.top <= other_submask_bottom):
            if compare_on_intersect:
                logging.warning('Notehead {0} intersecting other. Returning false.'.format(notehead.id))
                return False

        if notehead.bottom < other_submask_top:
            return False

        elif notehead.top > other_submask_bottom:
            return True

        else:
            raise NotationGraphError('Weird relative position of notehead'
                                     ' {0} and other {1}.'.format(notehead.id, other.id))

    def remove_vertex(self, node_id: int):
        self.remove_edges_for_vertex(node_id)
        node = self.__id_to_node_mapping[node_id]
        self.__nodes.remove(node)
        del self.__id_to_node_mapping[node_id]

    def remove_edge(self, from_id: int, to_id: int):
        if from_id not in self.__id_to_node_mapping:
            raise ValueError('Cannot remove edge from id {0}: not in graph!'
                             ''.format(from_id))
        if to_id not in self.__id_to_node_mapping:
            raise ValueError('Cannot remove edge to id {0}: not in graph!'
                             ''.format(to_id))

        from_node = self.__id_to_node_mapping[from_id]
        from_node.outlinks.remove(to_id)
        to_node = self.__id_to_node_mapping[to_id]
        to_node.inlinks.remove(from_id)

    def remove_edges_for_vertex(self, node_id: int):
        if node_id not in self.__id_to_node_mapping:
            raise ValueError('Cannot remove node with id {0}: not in graph!'
                             ''.format(node_id))
        node = self.__id_to_node_mapping[node_id]

        # Remove from inlinks and outlinks:
        for inlink in copy.deepcopy(node.inlinks):
            self.remove_edge(inlink, node_id)
        for outlink in copy.deepcopy(node.outlinks):
            self.remove_edge(node_id, outlink)

    def remove_classes(self, class_names: Iterable[str]):
        """Remove all vertices with these class names."""
        to_remove = [node.id for node in self.__nodes if node.class_name in class_names]
        for node_id in to_remove:
            self.remove_vertex(node_id)

    def remove_from_precedence(self, node_or_id: Union[Node, int]):
        """Bridge the precedence edges of the given object: each of its
        predecessors is linked to all of its descendants.
        If there are no predecessors or no descendants, the precedence
        edges are simply removed."""
        node_id = self.__to_id(node_or_id)
        node = self.__id_to_node_mapping[node_id]

        predecessors, descendants = [], []

        # Check if the node has at least some predecessors or descendants
        _has_predecessors = False
        if 'precedence_inlinks' in node.data:
            _has_predecessors = (len(node.data['precedence_inlinks']) > 0)
        if _has_predecessors:
            predecessors = copy.deepcopy(
                node.data['precedence_inlinks'])  # That damn iterator modification

        _has_descendants = False
        if 'precedence_outlinks' in node.data:
            _has_descendants = (len(node.data['precedence_outlinks']) > 0)
        if _has_descendants:
            descendants = copy.deepcopy(node.data['precedence_outlinks'])

        if (not _has_predecessors) and (not _has_descendants):
            return

        # Remove inlinks
        for predecessor_id in predecessors:
            predecessor = self.__id_to_node_mapping[predecessor_id]
            if 'precedence_outlinks' not in predecessor.data:
                raise ValueError(
                    'Predecessor {} of Node {} does not have precedence outlinks!'
                    ''.format(predecessor_id, node.id))
            if node.id not in predecessor.data['precedence_outlinks']:
                raise ValueError('Predecessor {} of Node {} does not have reciprocal outlink!'
                                 ''.format(predecessor_id, node.id))
            predecessor.data['precedence_outlinks'].remove(node.id)
            node.data['precedence_inlinks'].remove(predecessor_id)

        # Remove outlinks
        for descentant_id in descendants:
            descentant = self.__id_to_node_mapping[descentant_id]
            if 'precedence_inlinks' not in descentant.data:
                raise ValueError('Descendant {} of node {} does not have precedence inlinks!'
                                 ''.format(descentant_id, node.id))
            if node.id not in descentant.data['precedence_inlinks']:
                raise ValueError('Descendant {} of node {} does not have reciprocal inlink!'
                                 ''.format(descentant_id, node.id))
            descentant.data['precedence_inlinks'].remove(node.id)
            node.data['precedence_outlinks'].remove(descentant_id)

        # Bridge removed element
        for predecessor_id in predecessors:
            predecessor = self.__id_to_node_mapping[predecessor_id]
            for descentant_id in descendants:
                descentant = self.__id_to_node_mapping[descentant_id]
                if descentant_id not in predecessor.data['precedence_outlinks']:
                    predecessor.data['precedence_outlinks'].append(descentant_id)
                if predecessor_id not in descentant.data['precedence_inlinks']:
                    descentant.data['precedence_inlinks'].append(predecessor_id)

    def has_edge(self, from_id: int, to_id: int) -> bool:
        if from_id not in self.__id_to_node_mapping:
            logging.warning('Asking for object {}, which is not in graph.'.format(from_id))
        if to_id not in self.__id_to_node_mapping:
            logging.warning('Asking for object {}, which is not in graph.'.format(to_id))

        if to_id in self.__id_to_node_mapping[from_id].outlinks:
            if from_id in self.__id_to_node_mapping[to_id].inlinks:
                return True
            else:
                raise NotationGraphError('has_edge({}, {}): found {} in outlinks'
                                         ' of {}, but not {} in inlinks of {}!'
                                         ''.format(from_id, to_id, to_id, from_id, from_id, to_id))
        elif from_id in self.__id_to_node_mapping[to_id].inlinks:
            raise NotationGraphError('has_edge({}, {}): found {} in inlinks'
                                     ' of {}, but not {} in outlinks of {}!'
                                     ''.format(from_id, to_id, from_id, to_id, to_id, from_id))
        else:
            return False

    def add_edge(self, from_id: int, to_id: int):
        """Add an edge between the MuNGOs with ids ``fr --> to``.
        If the edge is already in the graph, warns and does nothing."""
        if from_id not in self.__id_to_node_mapping:
            raise NotationGraphError('Cannot remove edge from id {0}: not in graph!'.format(from_id))
        if to_id not in self.__id_to_node_mapping:
            raise NotationGraphError('Cannot remove edge to id {0}: not in graph!'.format(to_id))

        if to_id in self.__id_to_node_mapping[from_id].outlinks:
            if from_id in self.__id_to_node_mapping[to_id].inlinks:
                logging.info('Adding edge that is alredy in the graph: {} --> {}'
                             ' -- doing nothing'.format(from_id, to_id))
                return
            else:
                raise NotationGraphError('add_edge({}, {}): found {} in outlinks'
                                         ' of {}, but not {} in inlinks of {}!'
                                         ''.format(from_id, to_id, to_id, from_id, from_id, to_id))
        elif from_id in self.__id_to_node_mapping[to_id].inlinks:
            raise NotationGraphError('add_edge({}, {}): found {} in inlinks'
                                     ' of {}, but not {} in outlinks of {}!'
                                     ''.format(from_id, to_id, from_id, to_id, to_id, from_id))

        self.__id_to_node_mapping[from_id].outlinks.append(to_id)
        self.__id_to_node_mapping[to_id].inlinks.append(from_id)


##############################################################################


def group_staffs_into_systems(nodes: List[Node],
                              use_fallback_measure_separators: bool = True,
                              leftmost_measure_separators_only: bool = False) -> List[List[Node]]:
    """Returns a list of lists of ``staff`` Nodes
    grouped into systems. Uses the outer ``staff_grouping``
    symbols (or ``measure_separator``) symbols.

    Currently cannot deal with a situation where a system consists of
    interlocking staff groupings and measure separators, and cannot deal
    with system separator markings.

    :param nodes: The complete list of Nodes in the current
        document.

    :param use_fallback_measure_separators: If set and no staff groupings
        are found, will use measure separators instead to group
        staffs. The algorithm is to find the leftmost measure
        separator for each staff and use this set instead of staff
        groupings: measure separators also have outlinks to all
        staffs that they are relevant for.

    :param leftmost_measure_separators_only:

    :returns: A list of systems, where each system is a list of ``staff`` Nodes.
    """
    graph = NotationGraph(nodes)
    id_to_node_mapping = {c.id: c for c in nodes}
    staff_groups = [c for c in nodes
                    if c.class_name == 'staff_grouping']

    # Do not consider staffs that have no notehead or rest children.
    empty_staffs = [node for node in nodes if (node.class_name == 'staff') and
                    (len([inlink for inlink in node.inlinks
                          if ((id_to_node_mapping[inlink].class_name in _CONST.NOTEHEAD_CLASS_NAMES) or
                              (id_to_node_mapping[inlink].class_name in _CONST.REST_CLASS_NAMES))]) == 0)]
    print('Empty staffs: {0}'.format('\n'.join([str(node.id) for node in empty_staffs])))

    # There might also be non-empty staffs that are nevertheless
    # not covered by a staff grouping, only measure separators.

    if use_fallback_measure_separators:
        # Collect measure separators, sort them left to right
        measure_separators = [c for c in nodes if c.class_name in _CONST.MEASURE_SEPARATOR_CLASS_NAMES]
        measure_separators = sorted(measure_separators, key=operator.attrgetter('left'))
        # Use only the leftmost measure separator for each staff.
        staffs = [c for c in nodes if c.class_name in [_CONST.STAFF_CLASS_NAME]]

        if leftmost_measure_separators_only:
            leftmost_measure_separators = set()
            for staff in staffs:
                if staff in empty_staffs:
                    continue
                for m in measure_separators:
                    if graph.is_child_of(staff, m):
                        leftmost_measure_separators.add(m)
                        break
            staff_groups += leftmost_measure_separators
        else:
            staff_groups += measure_separators

    if len(staff_groups) != 0:
        staffs_per_group = {node.id: [id_to_node_mapping[i] for i in sorted(node.outlinks)
                                      if id_to_node_mapping[i].class_name == 'staff'] for node in staff_groups}
        # Build hierarchy of staff_grouping based on inclusion
        # between grouped staff sets.
        outer_staff_groups = []
        for staff_group in sorted(staff_groups, key=lambda c: c.left):
            sg_staffs = staffs_per_group[staff_group.id]
            is_outer = True
            for other_sg in staff_groups:
                if staff_group.id == other_sg.id:
                    continue
                other_sg_staffs = staffs_per_group[other_sg.id]
                if len([s for s in sg_staffs
                        if s not in other_sg_staffs]) == 0:
                    # If the staff groups are equal (can happen with
                    # a mixture of measure-separators and staff-groupings),
                    # only the leftmost should be an outer grouping.
                    if set(sg_staffs) == set(other_sg_staffs):
                        if other_sg in outer_staff_groups:
                            is_outer = False
                    else:
                        is_outer = False
            if is_outer:
                outer_staff_groups.append(staff_group)

        systems = [[c for c in nodes if (c.class_name == 'staff') and (c.id in staff_group.outlinks)] for staff_group in
                   outer_staff_groups]
    else:
        # Here we use the empty staff fallback
        systems = [[c] for c in nodes if (c.class_name == 'staff') and (c not in empty_staffs)]

    return systems


def group_by_staff(nodes: List[Node]) -> Dict[int, List[Node]]:
    """Returns one NotationGraph instance for each staff and its associated
    Nodes. "Associated" means:

    * the object is a descendant of the staff,
    * the object is an ancestor of the staff, or
    * the object is a descendant of an ancestor of the staff, *except*
      measure separators and staff groupings.
    """
    g = NotationGraph(nodes=nodes)

    staffs = [c for c in nodes if c.class_name == _CONST.STAFF_CLASS_NAME]
    objects_per_staff = dict()  # type: Dict[int, List[Node]]
    for staff in staffs:
        descendants = g.descendants(staff)
        ancestors = g.ancestors(staff)
        a_descendants = []
        for ancestor in ancestors:
            if ancestor.class_name in _CONST.SYSTEM_LEVEL_CLASS_NAMES:
                continue
            _ad = g.descendants(ancestor)
            a_descendants.extend(_ad)
        staff_related = set()
        for c in descendants + ancestors + a_descendants:
            staff_related.add(c)

        objects_per_staff[staff.id] = list(staff_related)

    return objects_per_staff


##############################################################################
# Graph search utilities

def find_related_staffs(query_nodes: List[Node], all_nodes: Union[NotationGraph, List[Node]],
                        with_stafflines: bool = True) -> List[Node]:
    """Find all staffs that are related to any of the nodes
    in question. Ignores whether these staffs are already within
    the list of ``query_nodes`` passed to the function.

    Finds all staffs that are ancestors or descendants of at least
    one of the query Nodes, and if ``with_stafflines`` is requested,
    all stafflines and staffspaces that are descendants of at least one
    of the related staffs as well.

    :param query_nodes: A list of Nodes for which we want
        to find related staffs. Subset of ``all_nodes``.

    :param all_nodes: A list of all the Nodes in the document
        (or directly a NotationGraph object). Assumes that the query
        Nodes are a subset of ``all_nodes``.

    :param with_stafflines: If set, will also return all stafflines
        and staffspaces related to the discovered staffs.

    :returns: List of staff (and, if requested, staffline/staffspace)
        Nodes that are relate to the query Nodes.
    """
    if not isinstance(all_nodes, NotationGraph):
        graph = NotationGraph(all_nodes)
    else:
        graph = all_nodes

    related_staffs = set()
    for c in query_nodes:
        desc_staffs = graph.descendants(c, class_filter=[_CONST.STAFF_CLASS_NAME])
        anc_staffs = graph.ancestors(c, class_filter=[_CONST.STAFF_CLASS_NAME])
        current_staffs = set(desc_staffs + anc_staffs)
        related_staffs = related_staffs.union(current_staffs)

    if with_stafflines:
        related_stafflines = set()
        for s in related_staffs:
            staffline_objs = graph.descendants(s, _CONST.STAFFLINE_CLASS_NAMES)
            related_stafflines = related_stafflines.union(set(staffline_objs))
        related_staffs = related_staffs.union(related_stafflines)

    return list(related_staffs)


##############################################################################
# Graph validation/fixing.
# An invariant of these methods should be that they never remove a correct
# edge. There is a known problem in this if a second stem is marked across
# staves: the beam orientation misfires.


def find_beams_incoherent_with_stems(nodes: List[Node]) -> List[List[Node]]:
    """Searches the graph for edges where a notehead is connected to a stem
    in one direction, but is connected to beams that are in the
    other direction.

    If a notehead has zero or more than one stem, it is ignored.

    :returns: A list of (notehead, beam) pairs such that the beam
        is not coherent with the stem direction for the notehead.
    """
    graph = NotationGraph(nodes)
    noteheads = [c for c in nodes if c.class_name in _CONST.NOTEHEAD_CLASS_NAMES]

    incoherent_pairs = []
    for notehead in noteheads:
        stems = graph.children(notehead, class_filter=['stem'])
        if len(stems) != 1:
            continue
        stem = stems[0]

        beams = graph.children(notehead, class_filter=['beam'])
        if len(beams) == 0:
            continue

        # Is the stem above the notehead, or not?
        # This is not trivial because of chords.
        is_stem_above = graph.is_stem_direction_above(notehead, stem)
        logging.info('IncoherentBeams: stem of {0} is above'.format(notehead.id))

        for beam in beams:
            try:
                is_beam_above = graph.is_symbol_above_notehead(notehead, beam)
            except NotationGraphError:
                logging.warning('IncoherentBeams: something is wrong in beam-notehead pair'
                                ' {0}, {1}'.format(beam.id, notehead.id))
                continue

            logging.info('IncoherentBeams: beam {0} of {1} is above'.format(beam.id, notehead.id))
            if is_stem_above != is_beam_above:
                incoherent_pairs.append([notehead, beam])

    return incoherent_pairs


# Leger lines often cause problems with autoparser.
# They should be always linked from noteheads in a consistent
# direction (from outside inwards to the staff).
# Also, no notehead should be connected to both a staffline/staffspace
# *AND* a leger line.

def find_leger_lines_with_noteheads_from_both_directions(nodes: List[Node]) -> List[Node]:
    """Looks for leger lines that have inlinks from noteheads
    on both sides. Returns a list of leger line Nodes."""
    graph = NotationGraph(nodes)

    problem_leger_lines = []

    for node in nodes:
        if node.class_name != _CONST.LEGER_LINE_CLASS_NAME:
            continue

        noteheads = graph.parents(node, class_filter=_CONST.NOTEHEAD_CLASS_NAMES)

        if len(noteheads) < 2:
            continue

        positions = [resolve_notehead_wrt_staffline(notehead, node) for notehead in noteheads]
        positions_not_on_staffline = [p for p in positions if p != 0]
        unique_positions = set(positions_not_on_staffline)
        if len(unique_positions) > 1:
            problem_leger_lines.append(node)

    return problem_leger_lines


def find_noteheads_with_leger_line_and_staff_conflict(nodes: List[Node]) -> List[Node]:
    """Find all noteheads that have a relationship both to a staffline
    or staffspace and to a leger line.

    Assumes (obviously) that staffline relationships have already been
    resolved. Useful in a workflow where autoparsing is applied *after*
    staff inference.
    """
    graph = NotationGraph(nodes)

    problem_noteheads = []

    for node in nodes:
        if node.class_name not in _CONST.NOTEHEAD_CLASS_NAMES:
            continue

        lls = graph.children(node, [_CONST.LEGER_LINE_CLASS_NAME])
        staff_objs = graph.children(node, _CONST.STAFFLINE_CLASS_NAMES)
        if lls and staff_objs:
            problem_noteheads.append(node)

    return problem_noteheads


def find_noteheads_on_staff_linked_to_leger_line(nodes: List[Node]) -> List[Node]:
    """Find all noteheads that are linked to a leger line,
    but at the same time intersect a staffline or lie
    entirely within a staffspace. These should be fixed
    by linking them to the corresponding staffline/staffspace,
    but the fixing operation should be in infer_staffline_relationships.

    This is the opposite of what ``resolve_leger_line_or_staffline_object()`` is doing.
    """
    graph = NotationGraph(nodes)
    problem_noteheads = []

    stafflines = sorted([c for c in nodes if c.class_name == 'staff_line'],
                        key=lambda x: x.top)
    staffspaces = sorted([c for c in nodes if c.class_name == 'staff_space'],
                         key=lambda x: x.top)

    for node in nodes:
        if node.class_name not in _CONST.NOTEHEAD_CLASS_NAMES:
            continue

        lls = graph.children(node, [_CONST.LEGER_LINE_CLASS_NAME])
        if len(lls) == 0:
            continue

        # Intersecting stafflines
        overlapped_stafflines = []
        for sl in stafflines:
            if node.overlaps(sl):
                overlapped_stafflines.append(sl)

        container_staffspaces = []
        for ss in staffspaces:
            if ss.contains(node):
                container_staffspaces.append(ss)

        if (len(overlapped_stafflines) + len(container_staffspaces)) > 0:
            problem_noteheads.append(node)

    return problem_noteheads


def find_misdirected_leger_line_edges(nodes: List[Node], retain_ll_for_disconnected_noteheads: bool = True) -> \
        List[List[Node]]:
    """Finds all edges that connect to leger lines, but do not
    lead in the direction of the staff.

    Silently assumes that all noteheads are connected to the correct staff.

    :param retain_ll_for_disconnected_noteheads:
        If the notehead would be left disconnected from all stafflines
        and staffspaces, retain its edges to its LLs -- it is better
        to get imperfect inference rather than for the PLAY button to fail.
    """
    graph = NotationGraph(nodes)

    misdirected_object_pairs = []

    for node in nodes:
        if node.class_name not in _CONST.NOTEHEAD_CLASS_NAMES:
            continue

        lls = graph.children(node, [_CONST.LEGER_LINE_CLASS_NAME])
        if not lls:
            continue

        staffs = graph.children(node, [_CONST.STAFF_CLASS_NAME])
        if not staffs:
            logging.warning('Notehead {0} not connected to any staff!'.format(node.id))
            continue
        staff = staffs[0]

        # Determine whether notehead is above or below staff.
        # Because of mistakes in notehead-ll edges, can actually be
        # *on* the staff. (If it is on a staffline, then the edge is
        # definitely wrong.)
        stafflines = sorted(graph.children(staff, [_CONST.STAFFLINE_CLASS_NAME]),
                            key=lambda x: x.top)
        p_top = resolve_notehead_wrt_staffline(node, stafflines[0])
        p_bottom = resolve_notehead_wrt_staffline(node, stafflines[-1])
        # Notehead actually located on the staff somewhere:
        # all of the LL rels. are false.
        if (p_top != p_bottom) or (p_top == 0) or (p_bottom == 0):
            for ll in lls:
                misdirected_object_pairs.append([node, ll])
            continue

        notehead_staff_direction = 1
        if p_bottom == -1:
            notehead_staff_direction = -1

        _current_misdirected_object_pairs = []
        for ll in lls:
            ll_direction = resolve_notehead_wrt_staffline(node, ll)
            if (ll_direction != 0) and (ll_direction != notehead_staff_direction):
                misdirected_object_pairs.append([node, ll])
                _current_misdirected_object_pairs.append([node, ll])

        if retain_ll_for_disconnected_noteheads:
            staffline_like_children = graph.children(node, class_filter=[_CONST.STAFFLINE_CLASS_NAME,
                                                                         _CONST.STAFFSPACE_CLASS_NAME,
                                                                         _CONST.LEGER_LINE_CLASS_NAME])
            # If all the notehead's links to staffline-like objects are scheduled to be discarded:
            if len(staffline_like_children) == len(_current_misdirected_object_pairs):
                # Remove them from the schedule
                misdirected_object_pairs = misdirected_object_pairs[:-len(_current_misdirected_object_pairs)]

    return misdirected_object_pairs


def resolve_leger_line_or_staffline_object(nodes: List[Node]):
    """If staff relationships are created before notehead to leger line
    relationships, then there will be noteheads on leger lines that
    are nevertheless connected to staffspaces. This function should be
    applied after both staffspace and leger line relationships have been
    inferred, to guess whether the notehead's relationship to the staff
    object should be discarded.

    Has no dependence on misdirected edge detection (handles this as a part
    of the conflict resolution).
    """
    graph = NotationGraph(nodes)

    for node in nodes:
        if node.class_name not in _CONST.NOTEHEAD_CLASS_NAMES:
            continue

        lls = graph.children(node, [_CONST.LEGER_LINE_CLASS_NAME])
        stafflines = graph.children(node, _CONST.STAFFLINE_CLASS_NAMES)
        staff = graph.children(node, _CONST.STAFF_CLASS_NAME)

        if len(lls) == 0:
            continue
        if len(stafflines) == 0:
            continue

        if len(staff) == 0:
            logging.warning('Notehead {0} not connected to any staff!'
                            ' Unable to resolve ll/staffline.'.format(node.id))
            continue

        # Multiple LLs: must check direction
        # Multiple stafflines: ???
        if len(stafflines) > 1:
            logging.warning('Notehead {0} is connected to multiple staffline'
                            ' objects!'.format(node.id))


##############################################################################

def group_by_measure(nodes: List[Node]):
    """Groups the objects into measures.
    Assumes the measures are consistent across staffs: no polytempi.

    If there are objects that span multiple measures, they are assigned
    to all the measures they intersect.

    If no measure separators are found, assumes everything belongs
    to one measure.

    :returns: A list of Node lists corresponding to measures. The list
        is ordered left-to-right.
    """
    graph = NotationGraph(nodes)
    logging.debug('Find measure separators.')

    measure_separators = [node for node in nodes if node.class_name in _CONST.MEASURE_SEPARATOR_CLASS_NAMES]

    if len(measure_separators) == 0:
        return nodes

    logging.debug('Order measure separators by precedence.')
    # Systems
    # measure seps. by system
    measure_separators = sorted(measure_separators, key=lambda m: m.left)

    logging.debug('Denote measure areas: bounding boxes and masks.')
    logging.debug('Assign objects to measures, based on overlap.')

    raise NotImplementedError()


##############################################################################
# Searching for MuNGOs that are contained within other MuNGOs
# and removing them safely from the MuNG.

def find_contained_nodes(nodes: List[Node], mask_threshold: float = 0.95):
    """Find all nodes that are contained within other nodes
    and not connected by an edge from container to contained.

    Does *NOT* check for transitive edges!"""
    graph = NotationGraph(nodes)

    # We should have some smarter indexing structure here, but since
    # we are just checking bounding boxes for candidates first,
    # it does not matter too much.

    nonstaff_nodes = [node for node in nodes if node.class_name not in _CONST.STAFF_CLASS_NAMES]

    contained_nodes = []
    for c1 in nonstaff_nodes:
        for c2 in nonstaff_nodes:
            if c1.id == c2.id:
                continue
            if c1.contains(c2):
                # Check mask overlap
                r, p, f = c1.compute_recall_precision_fscore_on_mask(c2)
                if r < mask_threshold:
                    continue
                if c2.id in c1.outlinks:
                    continue
                contained_nodes.append(c2)

    # Make unique
    return [c for c in set(contained_nodes)]


def remove_contained_nodes(nodes: List[Node], contained: List[Node]) -> List[Node]:
    """Removes ``contained`` Nodes from ``nodes`` so that the
    graph takes minimum damage.

    * Attachment edges of contained objects are removed.
    * For precedence edges, we link all precedence ancestors of a removed node
      to all its descendants.
    """
    # Operating on a copy. Inefficient, but safe.
    output_nodes = [copy.deepcopy(c) for c in nodes]

    # The nodes are then edited in-place by manipulating
    # the graph; hence we can then just return output_nodes.
    graph = NotationGraph(output_nodes)
    for c in contained:
        graph.remove_from_precedence(c.id)
    for c in contained:
        graph.remove_vertex(c.id)

    return graph.vertices


def resolve_notehead_wrt_staffline(notehead, staffline_or_leger_line) -> int:
    # type: (Node, Node) -> int
    """Resolves the relative vertical position of the notehead with respect
    to the given staff_line or legerLine object. Returns -1 if notehead
    is *below* staffline, 0 if notehead is *on* staffline, and 1 if notehead
    is *above* staffline."""
    ll = staffline_or_leger_line

    # Determining whether the notehead is on a leger
    # line or in the adjacent temp staffspace.
    # This uses a magic number, ON_STAFFLINE_RATIO_THRESHOLD.
    output_position = 0

    # Weird situation with notehead vertically *inside* bbox
    # of leger line (could happen with slanted LLs and very small
    # noteheads).
    if ll.top <= notehead.top <= notehead.bottom <= ll.bottom:
        output_position = 0

    # No vertical overlap between LL and notehead
    elif ll.top > notehead.bottom:
        output_position = 1
    elif notehead.top > ll.bottom:
        output_position = -1

    # Complicated situations: overlap
    else:
        # Notehead "around" leger line.
        if notehead.top < ll.top <= ll.bottom < notehead.bottom:
            dtop = ll.top - notehead.top
            dbottom = notehead.bottom - ll.bottom

            if min(dtop, dbottom) / max(dtop, dbottom) \
                    < _CONST.ON_STAFFLINE_RATIO_THRESHOLD:
                if dtop > dbottom:
                    output_position = 1
                else:
                    output_position = -1

        # Notehead interlaced with leger line, notehead on top
        elif notehead.top < ll.top <= notehead.bottom <= ll.bottom:
            output_position = 1

        # Notehead interlaced with leger line, leger line on top
        elif ll.top <= notehead.top <= ll.bottom < notehead.bottom:
            output_position = -1

        else:
            logging.warning('Strange notehead {0} vs. leger line {1}'
                            ' situation: bbox notehead {2}, LL {3}.'
                            ' Note that the output position is unusable;'
                            ' pleasre re-do this attachment manually.'
                            ''.format(notehead.id, ll.id,
                                      notehead.bounding_box,
                                      ll.bounding_box))
    return output_position


def is_notehead_on_line(notehead, line) -> bool:
    # type: (Node, Node) -> bool
    """Check whether given notehead is positioned on the line object."""
    if line.class_name not in _CONST.STAFFLINE_LIKE_CLASS_NAMES:
        raise ValueError('Cannot resolve relative position of notehead'
                         ' {0} to non-staffline-like object {1}'
                         ''.format(notehead.id, line.id))

    position = resolve_notehead_wrt_staffline(notehead, line)
    return position == 0



from setuptools import setup
import io
import logging
import os

import mung

here = os.path.abspath(os.path.dirname(__file__))


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


def get_long_description():
    readme = os.path.join(here, 'README.md')
    changes = os.path.join(here, 'CHANGES.md')

    if os.path.isfile(readme) and os.path.isfile(changes):
        long_description = read(readme, changes)
    else:
        logging.warning('Could not find README.md and CHANGES.md file'
                        ' in directory {0}. Contents:'
                        ' {1}'.format(here, os.listdir(here)))
        long_description = 'Tools for the Music Notation Graph representation of' \
                           ' music notation, used primarily for optical music' \
                           ' recognition. The MUSCIMA++ dataset uses this data' \
                           ' model. Supports export to MIDI. [README.md and' \
                           ' CHANGES.md not found]'
    return long_description


setup(
    name='mung',
    version=mung.__version__,
    url='https://mung.readthedocs.io',
    license='MIT Software License',
    author='Jan Hajič jr. and Alexander Pacha',
    install_requires=['numpy>=1.11.1',
                      'lxml>=3.6.4'],
    author_email='hajicj@ufal.mff.cuni.cz',
    description='Tools for the Music Notation Graph representation of music notation, used primarily for optical music recognition.',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    packages=['mung', 'mung2midi'],
    include_package_data=True,
    scripts=['scripts/add_staff_relationships.py',
             'scripts/add_staffline_symbols.py',
             'scripts/analyze_agreement.py',
             'scripts/analyze_annotations.py',
             'scripts/analyze_tracking_log.py',
             'scripts/baseline_process_symbols.py',
             'scripts/get_images_from_muscima.py',
             'scripts/infer_pitches.py',
             'scripts/strip_staffline_symbols.py',
             ],
    platforms='any',
    test_suite='test.test_mung',
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: Web Environment',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
    ],
)



"""This module implements functions for reading and writing the data formats used by MUSCIMA++.

Data formats
============

All MUSCIMA++ data is stored as XML, in ``<Node>`` elements.
These are grouped into ``<Nodes>`` elements, which are
the top-level elements in the ``*.xml`` dataset files.

The list of object classes used in the dataset is also stored as XML,
in ``<NodeClass>`` elements (within a ``<NodeClasses>``
element).

Node
----------

To read a Node list file (in this case, a test data file):

    >>> from mung.io import read_nodes_from_file
    >>> import os
    >>> file = os.path.join(os.path.dirname(__file__), '../test/test_data/01_basic.xml')
    >>> nodes = read_nodes_from_file(file)

The ``Node`` string representation is a XML object::

    <Node xml:id="MUSCIMA-pp_1.0___CVC-MUSCIMA_W-01_N-10_D-ideal___25">
      <Id>25</Id>
      <ClassName>grace-notehead-full</ClassName>
      <Top>119</Top>
      <Left>413</Left>
      <Width>16</Width>
      <Height>6</Height>
      <Mask>1:5 0:11 (...) 1:4 0:6 1:5 0:1</Mask>
      <Outlinks>12 24 26</Outlinks>
      <Inlinks>13</Inlinks>
    </Node>

The Nodes are themselves kept as a list::

    <Nodes>
      <Node> ... </Node>
      <Node> ... </Node>
    </Nodes>

Parsing is only implemented for files that consist of a single
``<Nodes>``.

Additional information
^^^^^^^^^^^^^^^^^^^^^^

.. caution::

    This part may easily be deprecated.

Arbitrary data can be added to the Node using the optional
``<Data>`` element. It should encode a dictionary of additional
information about the Node that may only apply to a subset
of Nodes (this facultativeness is what distinguishes the
purpose of the ``<Data>`` element from just subclassing ``Node``).

For example, encoding the pitch, duration and precedence information
about a notehead could look like this::

    <Node>
        ...
        <Data>
            <DataItem key="pitch_step" type="str">D</DataItem>
            <DataItem key="pitch_modification" type="int">1</DataItem>
            <DataItem key="pitch_octave" type="int">4</DataItem>
            <DataItem key="midi_pitch_code" type="int">63</DataItem>
            <DataItem key="midi_duration" type="int">128</DataItem>
            <DataItem key="precedence_inlinks" type="list[int]">23 24 25</DataItem>
            <DataItem key="precedence_outlinks" type="list[int]">27</DataItem>
        </Data>
    </Node

The ``Node`` will then contain in its ``data`` attribute
the dictionary::

    self.data = {'pitch_step': 'D',
                 'pitch_modification': 1,
                 'pitch_octave': 4,
                 'midi_pitch_code': 63,
                 'midi_pitch_duration': 128,
                 'precedence_inlinks': [23, 24, 25],
                 'precedence_outlinks': [27]}


This is also a basic mechanism to allow you to subclass
Node with extra attributes without having to re-implement
parsing and export.

.. warning::

    Do not misuse this! The ``<Data>`` mechanism is primarily
    intended to encode extra information for MUSCIMarker to
    display.


Individual elements of a ``<Node>``
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

* ``<Id>`` is the unique integer ID of the Node inside this document
* ``<ClassName>`` is the name of the object's class (such as
  ``noteheadFull``, ``beam``, ``numeral3``, etc.).
* ``<Top>`` is the vertical coordinate of the upper left corner of the object's
  bounding box.
* ``<Left>`` is the horizontal coordinate of the upper left corner of
  the object's bounding box.
* ``<Width>``: the amount of rows that the Node spans.
* ``<Height>``: the amount of columns that the Node spans.
* ``<Mask>``: a run-length-encoded binary (0/1) array that denotes the area
  within the Node's bounding box (specified by ``top``, ``left``,
  ``height`` and ``width``) that the Node actually occupies. If
  the mask is not given, the object is understood to occupy the entire
  bounding box. For the representation, see Implementation notes
  below.
* ``<Inlinks>``: whitespace-separate ``id`` list, representing Nodes
  **from** which a relationship leads to this Node. (Relationships are
  directed edges, forming a directed graph of Nodes.) The ids are
  valid in the same scope as the Node's ``id``: don't mix
  Nodes from multiple scopes (e.g., multiple documents)!
  If you are using Nodes from multiple documents at the same
  time, make sure to check against the ``unique_id``s.
* ``<Outlinks>``: whitespace-separate ``id`` list, representing Nodes
  **to** which a relationship leads to this Node. (Relationships are
  directed edges, forming a directed graph of Nodes.) The ids are
  valid in the same scope as the Node's ``id``: don't mix
  Nodes from multiple scopes (e.g., multiple documents)!
  If you are using Nodes from multiple documents at the same
  time, make sure to check against the ``unique_id``s.
* ``<Data>``: a list of ``<DataItem>`` elements. The elements have
  two attributes: ``key``, and ``type``. The ``key`` is what the item
  should be called in the ``data`` dict of the loaded Node.
  The ``type`` attribute encodes the Python type of the item and gets
  applied to the text of the ``<DataItem>`` to produce the value.
  Currently supported types are ``int``, ``float``, and ``str``,
  and ``list[int]``, ``list[float]`` and ``list[str]``. The lists
  are whitespace-separated.

The parser function provided for Nodes does *not* check against
the presence of other elements. You can extend Nodes for your
own purposes -- but you will have to implement parsing.

Implementation notes on the mask
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The mask is a numpy array that will be saved using run-length encoding.
The numpy array is first flattened, then runs of successive 0's and 1's
are encoded as e.g. ``0:10 `` for a run of 10 zeros.

How much space does this take?

Objects tend to be relatively convex, so after flattening, we can expect
more or less two runs per row (flattening is done in ``C`` order). Because
each run takes (approximately) 5 characters, each mask takes roughly ``5 * n_rows``
bytes to encode. This makes it efficient for objects wider than 5 pixels, with
a compression ratio approximately ``n_cols / 5``.
(Also, the numpy array needs to be made C-contiguous for that, which
explains the `NODE_MASK_ORDER='C'` hack in `set_mask()`.)


NodeClass
---------------

This is what a single NodeClass element might look like::

    <NodeClass>
        <Id>1</Id>
        <Name>notehead-empty</Name>
        <GroupName>note-primitive/notehead-empty</GroupName>
        <Color>#FF7566</Color>
    </NodeClass>

See e.g. ``test/test_data/mff-muscima-classes-annot.xml``,
which is incidentally the real NodeClass list used
for annotating MUSCIMA++.

Similarly to a ``<Nodes>``, the ``<NodeClass>``
elements are organized inside a ``<NodeClasses>``::

   <NodeClasses>
      <NodeClass> ... </NodeClass>
      <NodeClass> ... </NodeClass>
   </NodeClasses>

The :class:`NodeClass` represents one possible :class:`Node`
symbol class, such as a notehead or a time signature. Aside from defining
the "vocabulary" of available object classes for annotation, it also contains
some information about how objects of the given class should
be displayed in the MUSCIMarker annotation software (ordering
related object classes together in menus, implementing a sensible
color scheme, etc.). There is nothing interesting about this class,
we pulled it into the ``mung`` package because the object
grammar (i.e. which relationships are allowed and which are not)
depends on having NodeClass object as its "vocabulary",
and you will probably want to manipulate the data somehow based
on the objects' relationships (like reassembling notes from notation
primitives: notehead plus stem plus flags...), and the grammar
file is a reference for doing that.

"""
import logging
import os

import collections
from typing import List, Tuple

from lxml import etree

from mung.node import Node
from mung.node_class import NodeClass


def read_nodes_from_file(filename: str) -> List[Node]:
    """From a xml file with a Nodes as the top element, parse
    a list of nodes. (See ``Node`` class documentation
    for a description of the XMl format.)

    Let's test whether the parsing function works:

    >>> test_data_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)),
    ...                              'test', 'test_data')
    >>> file = os.path.join(test_data_dir, '01_basic.xml')
    >>> nodes = read_nodes_from_file(file)
    >>> len(nodes)
    48

    Let's also test the ``data`` attribute:
    >>> file_with_data = os.path.join(test_data_dir, '01_basic_binary.xml')
    >>> nodes = read_nodes_from_file(file_with_data)
    >>> nodes[0].data['pitch_step']
    'G'
    >>> nodes[0].data['midi_pitch_code']
    79
    >>> nodes[0].data['precedence_outlinks']
    [8, 17]
    >>> nodes[0].dataset
    'testdataset'
    >>> nodes[0].document
    '01_basic_binary'

    :returns: A list of ``Node``s.
    """
    tree = etree.parse(filename)
    root = tree.getroot()
    logging.debug('XML parsed.')
    nodes = []
    dataset = root.attrib['dataset']
    document = root.attrib['document']

    for i, node in enumerate(root.iter('Node')):
        ######################################################
        logging.debug('Parsing Node {0}'.format(i))

        node_id = int(float(node.findall('Id')[0].text))
        class_name = node.findall('ClassName')[0].text
        top = int(node.findall('Top')[0].text)
        left = int(node.findall('Left')[0].text)
        width = int(node.findall('Width')[0].text)
        height = int(node.findall('Height')[0].text)

        #################################
        # Parsing the graph structure (Can deal with missing Inlinks/Outlinks)
        inlinks = []
        i_s = node.findall('Inlinks')
        if len(i_s) > 0:
            i_s_text = node.findall('Inlinks')[0].text
            if i_s_text is not None:  # Zero-length links
                inlinks = list(map(int, i_s_text.split(' ')))

        outlinks = []
        o_s = node.findall('Outlinks')
        if len(o_s) > 0:
            o_s_text = node.findall('Outlinks')[0].text
            if o_s_text is not None:
                outlinks = list(map(int, o_s_text.split(' ')))

        #################################
        data = node.findall('Data')
        data_dict = None
        if len(data) > 0:
            data = data[0]
            data_dict = {}
            for data_item in data.findall('DataItem'):
                key = data_item.get('key')
                value_type = data_item.get('type')
                value = data_item.text

                # logging.debug('Creating data entry: key={0}, type={1},'
                #              ' value={2}'.format(key, value_type, value))

                if value_type == 'int':
                    value = int(value)
                elif value_type == 'float':
                    value = float(value)
                elif value_type.startswith('list'):
                    if value is None:
                        value = []
                    else:
                        vt_factory = str
                        if value_type.endswith('[int]'):
                            vt_factory = int
                        elif value_type.endswith('[float]'):
                            vt_factory = float
                        value = list(map(vt_factory, value.split()))

                data_dict[key] = value

        #################################
        # Create the object.
        new_node = Node(id_=node_id,
                        class_name=class_name,
                        top=top,
                        left=left,
                        width=width,
                        height=height,
                        inlinks=inlinks,
                        outlinks=outlinks,
                        dataset=dataset,
                        document=document,
                        data=data_dict)

        #################################
        # Add mask.
        # We do this only after the Node has been created,
        # to make sure that the width & height used to reshape
        # the flattened mask reflects what is in the Node.
        mask = None
        mask_elements = node.findall('Mask')
        if len(mask_elements) > 0:
            mask = Node.decode_mask(mask_elements[0].text, shape=(new_node.height, new_node.width))
        new_node.set_mask(mask)
        nodes.append(new_node)

    logging.debug('Nodes loaded.')

    if not validate_nodes_graph_structure(nodes):
        raise ValueError('Invalid Node graph structure! Check warnings'
                         ' in log for the individual errors.')

    return nodes


def validate_nodes_graph_structure(nodes: List[Node]) -> bool:
    """Check that the graph defined by the ``inlinks`` and ``outlinks``
    in the given list of Nodes is valid: no relationships
    leading from or to objects with non-existent ``id``s.

    Can deal with ``Nodes`` coming from a combination
    of documents, through the Node ``document`` property.
    Warns about documents which are found inconsistent.

    :param nodes: A list of :class:`Node` instances.

    :returns: ``True`` if graph is valid, ``False`` otherwise.
    """
    # Split into lists by document
    nodes_by_document = collections.defaultdict(list)
    for node in nodes:
        nodes_by_document[node.document].append(node)

    is_valid = True
    for doc, document_nodes in list(nodes_by_document.items()):
        document_is_valid = validate_document_graph_structure(document_nodes)
        if not document_is_valid:
            logging.warning('Document {0} has invalid Node graph!'.format(doc))
            is_valid = False
    return is_valid


def validate_document_graph_structure(nodes: List[Node]) -> bool:
    """Check that the graph defined by the ``inlinks`` and ``outlinks``
    in the given list of Nodes is valid: no relationships
    leading from or to objects with non-existent ``id``s.

    Checks that all the Nodes come from one document. (Raises
    a ``ValueError`` otherwise.)

    :param nodes: A list of :class:`Node` instances.

    :returns: ``True`` if graph is valid, ``False`` otherwise.
    """
    docs = [node.document for node in nodes]
    if len(set(docs)) != 1:
        raise ValueError('Got Nodes from multiple documents!')

    is_valid = True
    node_ids = frozenset([node.id for node in nodes])
    for c in nodes:
        inlinks = c.inlinks
        for i in inlinks:
            if i not in node_ids:
                logging.warning('Invalid graph structure in NodeList:'
                                ' object {0} has inlink from non-existent'
                                ' object {1}'.format(c, i))
                is_valid = False

        outlinks = c.outlinks
        for o in outlinks:
            if o not in node_ids:
                logging.warning('Invalid graph structure in NodeList:'
                                ' object {0} has outlink to non-existent'
                                ' object {1}'.format(c, o))
                is_valid = False

    return is_valid


def get_edges(nodes: List[Node], validate: bool = True) -> List[Tuple[int, int]]:
    """Collects the inlink/outlink Node graph
    and returns it as a list of ``(from, to)`` edges.

    :param nodes: A list of Node instances.
        All are expected to be within one document.

    :param validate: If set, will raise a ValueError
        if the graph defined by the Nodes is
        invalid.

    :returns: A list of ``(from, to)`` id pairs
        that represent edges in the Node graph.
    """
    if validate:
        validate_nodes_graph_structure(nodes)

    edges = []
    for c in nodes:
        for o in c.outlinks:
            edges.append((c.id, o))
    return edges


def export_node_list(nodes: List[Node], document: str = None, dataset: str = None) -> str:
    """Writes the Node data as a XML string. Does not write
    to a file -- use ``with open(output_file) as out_stream:`` etc.

    """
    # This is the data string, the rest is formalities
    nodes_string = '\n'.join([str(c) for c in nodes])

    lines = list()
    lines.append('<?xml version="1.0" encoding="utf-8"?>')
    lines.append('<Nodes dataset={0} document={1}'
                 ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
                 ' xmlns:xsd="http://www.w3.org/2001/XMLSchema"'
                 ' xsi:noNamespaceSchemaLocation="CVC-MUSCIMA_Schema.xsd">'.format(dataset, document))
    lines.append(nodes_string)
    lines.append('</Nodes>')
    return '\n'.join(lines)


##############################################################################
# Parsing NodeClass lists, mostly for grammars.

def parse_node_classes(filename: str) -> List[NodeClass]:
    """ Extract the list of :class:`NodeClass` objects from
        an xml file with a NodeClasses as the top element and NodeClass children.
    """
    tree = etree.parse(filename)
    node_classes_xml = tree.getroot()
    node_classes = []
    for node_class_xml in node_classes_xml:
        if node_class_xml.tag != "NodeClass":
            continue
        node_class = NodeClass(class_id=int(node_class_xml.findall('Id')[0].text),
                               name=node_class_xml.findall('Name')[0].text,
                               group_name=node_class_xml.findall('GroupName')[0].text,
                               color=node_class_xml.findall('Color')[0].text)
        node_classes.append(node_class)
    return node_classes


def export_nodeclass_list(node_classes: List[NodeClass]) -> str:
    """Writes the Node data as a XML string. Does not write
    to a file -- use ``with open(output_file) as out_stream:`` etc.
    """
    # This is the data string, the rest is formalities
    node_classes_string = '\n'.join([str(c) for c in node_classes])

    lines = list()

    lines.append('<?xml version="1.0" encoding="utf-8"?>')
    lines.append('<NodeClasses noNamespaceSchema="mff-muscima-mlclasses.xsd">')
    lines.append(node_classes_string)
    lines.append('</NodeClasses>')
    return '\n'.join(lines)



from mung.grammar import DependencyGrammar
from mung.io import parse_node_classes

import os
import unittest


class GrammarTest(unittest.TestCase):
    def test_parse_grammar(self):
        filepath = os.path.dirname(
            os.path.dirname(__file__)) + u'/test/test_data/mff-muscima-classes-annot.deprules'
        node_classes_path = os.path.dirname(
            os.path.dirname(__file__)) + u'/test/test_data/mff-muscima-classes-annot.xml'
        node_classes = parse_node_classes(node_classes_path)
        node_classes_dict = {node_class.name for node_class in node_classes}
        dependency_graph = DependencyGrammar(grammar_filename=filepath, alphabet=node_classes_dict)
        self.assertEqual(646, len(dependency_graph.rules))


if __name__ == '__main__':
    unittest.main()
